#!/usr/bin/python3
"""
Copyright (C) 2023 Michael Ablassmeier <abi@grinser.de>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import logging
from argparse import Namespace
from typing import List
from libvirt import virDomain, libvirtError
from libvirtnbdbackup import virt
from libvirtnbdbackup.virt.client import DomainDisk
from libvirtnbdbackup.virt.exceptions import startBackupFailed


def start(
    args: Namespace,
    virtClient: virt.client,
    domObj: virDomain,
    disks: List[DomainDisk],
) -> bool:
    """Start backup job via libvirt API"""

    paused: bool = False

    if args.pause is True and args.start_domain is False:
        try:
            domObj.suspend()
            paused = True
            logging.info("Paused virtual machine.")
        except libvirtError as e:
            logging.warning("Attempting to pause VM failed: [%s]", e)

    try:
        logging.info("Starting backup job.")
        virtClient.startBackup(
            args,
            domObj,
            disks,
        )
        logging.debug("Backup job started.")
        return True
    except startBackupFailed as e:
        logging.error(e)
    finally:
        if paused is True:
            try:
                domObj.resume()
                logging.info("Resumed virtual machine.")
            except libvirtError as e:
                logging.warning("Attempting to resume VM failed: [%s]", e)

    return False



"""
Copyright (C) 2026  Michael Ablassmeier <abi@grinser.de>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import json
import time
import base64
import logging
import libvirt
import libvirt_qemu


def Exec(
    domObj: libvirt.virDomain,
    command: str,
    args: list[str],
    timeout: int = 30,
) -> str:
    """Execute command within VM using guest-agent"""

    def agent_cmd(payload: dict) -> dict:
        body = json.dumps(payload)
        # Some bindings expose this on virDomain, others via libvirt_qemu.
        if hasattr(domObj, "qemuAgentCommand"):
            raw = domObj.qemuAgentCommand(body, timeout, 0)
        else:
            raw = libvirt_qemu.qemuAgentCommand(domObj, body, timeout, 0)
        return json.loads(raw)

    # Start process in guest.
    payload = {
        "execute": "guest-exec",
        "arguments": {
            "path": f"{command}",
            "arg": args,
            "capture-output": True,
        },
    }
    result = agent_cmd(payload)
    pid = result["return"]["pid"]

    stdout: str = ""
    stderr: str = ""

    deadline = time.time() + timeout
    while time.time() < deadline:
        status_req = {
            "execute": "guest-exec-status",
            "arguments": {"pid": pid},
        }
        status = agent_cmd(status_req)["return"]
        if status.get("exited"):
            out_data = status.get("out-data")
            err_data = status.get("err-data")
            if out_data:
                stdout = base64.b64decode(out_data).decode(errors="replace")
                logging.debug(stdout)
            if err_data:
                stderr = base64.b64decode(err_data).decode(errors="replace")
                logging.debug(stderr)

            if status.get("exitcode", 1) == 1:
                raise RuntimeError(
                    f"Executing guest command failed: [{stdout}] [{stderr}]"
                )

            return stdout

        time.sleep(0.5)

    raise TimeoutError(f"Timed out waiting for command in guest (pid={pid})")



"""
Copyright (C) 2023  Michael Ablassmeier <abi@grinser.de>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import os
import sys
import zipfile
import logging
import time
from typing import IO, Tuple
from libvirtnbdbackup.output import exceptions
from libvirtnbdbackup.output.target.directory import Directory

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal

log = logging.getLogger("zip")


class Zip:
    """Backup to zip file"""

    def __init__(self) -> None:
        self.zipStream: zipfile.ZipFile
        self.zipFileStream: IO[bytes]

        log.info("Writing zip file stream to stdout")
        try:
            # pylint: disable=consider-using-with
            self.zipStream = zipfile.ZipFile(sys.stdout.buffer, "x", zipfile.ZIP_STORED)
        except zipfile.error as e:
            raise exceptions.OutputOpenException(f"Failed to open zip file: {e}") from e

    def create(self, targetDir) -> None:
        """Create wrapper"""
        log.debug("Create: %s", targetDir)
        Directory().create(targetDir)

    def open(self, fileName: str, mode: Literal["w"] = "w") -> IO[bytes]:
        """Open wrapper"""
        zipFile = zipfile.ZipInfo(
            filename=os.path.basename(fileName),
        )
        dateTime: time.struct_time = time.localtime(time.time())
        timeStamp: Tuple[int, int, int, int, int, int] = (
            dateTime.tm_year,
            dateTime.tm_mon,
            dateTime.tm_mday,
            dateTime.tm_hour,
            dateTime.tm_min,
            dateTime.tm_sec,
        )
        zipFile.date_time = timeStamp
        zipFile.compress_type = zipfile.ZIP_STORED

        try:
            # pylint: disable=consider-using-with
            self.zipFileStream = self.zipStream.open(zipFile, mode, force_zip64=True)
            return self.zipFileStream
        except zipfile.error as e:
            raise exceptions.OutputOpenException(
                f"Failed to open zip stream: {e}"
            ) from e

        return False

    def truncate(self, size: int) -> None:
        """Truncate target file"""
        raise RuntimeError("Not implemented")

    def write(self, data: bytes) -> int:
        """Write wrapper"""
        return self.zipFileStream.write(data)

    def close(self) -> None:
        """Close wrapper"""
        log.debug("Close file")
        self.zipFileStream.close()

    def checksum(self) -> None:
        """Checksum: not implemented for zip file"""
        return



"""
Copyright (C) 2023  Michael Ablassmeier <abi@grinser.de>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import logging
from lxml.etree import _Element
from lxml import etree as ElementTree

log = logging.getLogger()


def asTree(vmConfig: str) -> _Element:
    """Return Etree element for vm config"""
    return ElementTree.fromstring(vmConfig)


def indent(top: _Element) -> str:
    """Indent xml output for debug log"""
    try:
        ElementTree.indent(top)
    except ElementTree.ParseError as errmsg:
        log.debug("Failed to parse xml: [%s]", errmsg)
    except AttributeError:
        # older ElementTree versions dont have the
        # indent method, skip silently and use
        # non formatted string
        pass

    xml = ElementTree.tostring(top).decode()
    log.debug("\n%s", xml)

    return xml



"""
Exceptions
"""


class virtHelperError(Exception):
    """Errors during libvirt helper"""


class domainNotFound(virtHelperError):
    """Can't find domain"""


class connectionFailed(virtHelperError):
    """Can't connect libvirtd domain"""


class startBackupFailed(virtHelperError):
    """Can't start backup operation"""



"""
Copyright (C) 2023  Michael Ablassmeier <abi@grinser.de>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import json
import logging
import tempfile
import subprocess
from typing import List
from argparse import Namespace

from libvirtnbdbackup.ssh.exceptions import sshError
from libvirtnbdbackup.objects import processInfo
from libvirtnbdbackup.qemu import command
from libvirtnbdbackup.virt.client import DomainDisk

log = logging.getLogger(__name__)


class util:
    """Wrapper for qemu executables"""

    def __init__(self, exportName: str) -> None:
        self.exportName = exportName
        self.restoreCmd = [
            "qemu-nbd",
            "--fork",
            "--discard=unmap",
            "-x",
            f"{self.exportName}",
        ]

    @staticmethod
    def _getcompress(args: Namespace, targetFile: str) -> List:
        """Check if --compress option is set and if so, use
        qemu's compress driver during data write'"""
        cmd = []
        if args.compress is not True:
            cmd.append(targetFile)
            cmd.append("--format=qcow2")
        else:
            cmd.append("--image-opts")
            cmd.append(
                f"driver=compress,file.driver=qcow2,file.file.driver=file,file.file.filename={targetFile}"  # pylint: disable=line-too-long
            )

        return cmd

    @staticmethod
    def map(cType, context: str) -> str:
        """Read extent map using nbdinfo utility"""
        cmd = f"nbdinfo --json --map={context} '{cType.uri}'"
        log.debug("Starting CMD: [%s]", cmd)
        extentMap = subprocess.run(
            cmd,
            shell=True,
            check=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )

        return json.loads(extentMap.stdout)

    @staticmethod
    def create(  # pylint: disable=R0913: disable=R0913
        args: Namespace,
        targetFile: str,
        fileSize: int,
        diskFormat: str,
        qcowOptions: list,
        sshClient=None,
    ) -> processInfo:
        """Create the target qcow image"""
        fileParam = f"{targetFile}"
        if sshClient:
            fileParam = f"'{targetFile}'"

        cmd = [
            "qemu-img",
            "create",
            "-f",
            f"{diskFormat}",
            fileParam,
            "-o",
            f"size={fileSize}",
        ]

        if args.preallocate:
            cmd.append("-o")
            cmd.append("preallocation=full")

        if qcowOptions:
            cmd = cmd + qcowOptions

        if not sshClient:
            return command.run(cmd)

        return sshClient.run(" ".join(cmd))

    def info(self, targetFile: str, sshClient=None) -> processInfo:
        """Return qemu image information"""
        fileParam = f"{targetFile}"
        if sshClient:
            fileParam = f"'{targetFile}'"

        cmd = [
            "qemu-img",
            "info",
            fileParam,
            "--output",
            "json",
            "--force-share",
        ]
        if not sshClient:
            return command.run(cmd, toPipe=True)

        return sshClient.run(" ".join(cmd))

    def startRestoreNbdServer(self, args: Namespace, targetFile: str) -> processInfo:
        """Start local nbd server process for restore operation"""
        pidFile = self._gt("qemu-nbd", ".pid")
        cmd = self.restoreCmd
        cmd = cmd + self._getcompress(args, targetFile)
        cmd.append("-k")
        cmd.append(args.socketfile)
        cmd.append("--pid-file")
        cmd.append(pidFile)
        return command.run(cmd, pidFile=pidFile)

    @staticmethod
    def _gt(prefix: str, suffix: str, delete: bool = False) -> str:
        """Create named temporary file."""
        with tempfile.NamedTemporaryFile(
            delete=delete, prefix=prefix, suffix=suffix
        ) as tf1:
            return tf1.name

    @staticmethod
    def _addTls(cmd: List[str], certpath: str) -> None:
        """Add required tls related options to qemu-nbd command
        line."""
        cmd.append("--object")
        cmd.append(
            f"tls-creds-x509,id=tls0,endpoint=server,dir={certpath},verify-peer=false"
        )
        cmd.append("--tls-creds tls0")

    def startRemoteRestoreNbdServer(
        self, args: Namespace, targetFile: str
    ) -> processInfo:
        """Start nbd server process remotely over ssh for restore operation"""
        pidFile = self._gt("qemu-nbd-restore", ".pid")
        logFile = self._gt("qemu-nbd-restore", ".log")
        cmd = self.restoreCmd
        cmd = cmd + self._getcompress(args, targetFile)
        cmd.append("-p")
        cmd.append(f"{args.nbd_port}")
        cmd.append("--pid-file")
        cmd.append(pidFile)
        if args.tls is True:
            self._addTls(cmd, args.tls_cert)
        if args.nbd_ip != "":
            cmd.append("-b")
            cmd.append(args.nbd_ip)
        cmd.append(f"> {logFile} 2>&1")
        logging.debug(cmd)
        try:
            return args.sshClient.run(" ".join(cmd), pidFile, logFile)
        except sshError:
            log.error("Executing command failed: check [%s] for errors.", logFile)
            raise

    def startNbdkitProcess(
        self, args: Namespace, nbdkitModule: str, blockMap, fullImage: str
    ) -> processInfo:
        """Execute nbdkit process for virtnbdmap"""
        debug = "0"
        hexdump = "0"
        pidFile = self._gt("nbdkit", ".pid")
        if args.verbose:
            debug = "1"
        if args.hexdump:
            hexdump = "1"
            debug = "1"
        cmd = [
            "nbdkit",
            "--pidfile",
            f"{pidFile}",
            "-i",
            f"{args.listen_address}",
            "-p",
            f"{args.listen_port}",
            "-e",
            f"{self.exportName}",
            "--filter=blocksize",
            "--filter=cow",
            "-v",
            "python",
            f"{nbdkitModule}",
            f"maxlen={args.blocksize}",
            f"blockmap={blockMap}",
            f"disk={fullImage}",
            f"debug={debug}",
            f"hexdump={hexdump}",
            "-t",
            f"{args.threads}",
        ]
        return command.run(cmd, pidFile=pidFile)

    def startBackupNbdServer(
        self, diskFormat: str, diskFile: str, socketFile: str, bitMap: str
    ) -> processInfo:
        """Start nbd server process for offline backup operation"""
        bitmapOpt = "--"
        if bitMap != "":
            bitmapOpt = f"--bitmap={bitMap}"

        pidFile = f"{socketFile}.pid"
        cmd = [
            "qemu-nbd",
            "-r",
            f"--format={diskFormat}",
            "-x",
            f"{self.exportName}",
            f"{diskFile}",
            "-k",
            f"{socketFile}",
            "-t",
            "-e 2",
            "--fork",
            "--detect-zeroes=on",
            f"--pid-file={pidFile}",
            bitmapOpt,
        ]
        return command.run(cmd, pidFile=pidFile)

    def startRemoteBackupNbdServer(
        self, args: Namespace, disk: DomainDisk, bitMap: str, port: int
    ) -> processInfo:
        """Start nbd server process remotely over ssh for restore operation"""
        pidFile = self._gt("qemu-nbd-backup", ".pid")
        logFile = self._gt("qemu-nbd-backup", ".log")
        cmd = [
            "qemu-nbd",
            "-r",
            f"--format={disk.format}",
            "-x",
            f"{self.exportName}",
            f"'{disk.path}'",
            "-p",
            f"{port}",
            "--pid-file",
            f"{pidFile}",
            "--fork",
        ]
        if args.nbd_ip != "":
            cmd.append("-b")
            cmd.append(args.nbd_ip)

        if bitMap != "":
            cmd.append(f"--bitmap={bitMap}")

        if args.tls is True:
            self._addTls(cmd, args.tls_cert)
        cmd.append(f"> {logFile} 2>&1")
        try:
            return args.sshClient.run(" ".join(cmd), pidFile, logFile)
        except sshError:
            log.error("Executing command failed: check [%s] for errors.", logFile)
            raise

    def disconnect(self, device: str) -> processInfo:
        """Disconnect device"""
        log.info("Disconnecting device [%s]", device)
        cmd = ["qemu-nbd", "-d", f"{device}"]
        return command.run(cmd)



"""
Exceptions
"""


class StreamFormatException(Exception):
    """Wrong metadata header"""


class MetaHeaderFormatException(StreamFormatException):
    """Wrong metadata header"""


class BlockFormatException(StreamFormatException):
    """Wrong metadata header"""


class FrameformatException(StreamFormatException):
    """Frame Format is wrong"""



"""
Copyright (C) 2023  Michael Ablassmeier <abi@grinser.de>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from typing import List, Any, Tuple, IO, Union
from nbd import Error as nbdError
from libvirtnbdbackup import block
from libvirtnbdbackup import lz4
from libvirtnbdbackup.exceptions import DiskBackupFailed

# pylint: disable=too-many-arguments


def write(
    writer: IO[Any], blk, nbdCon, btype: str, compress: Union[bool, int], pbar
) -> Tuple[int, List[int]]:
    """During extent processing, consecutive blocks with
    the same type(data or zeroed) are unified into one big chunk.
    This helps to reduce requests to the NBD Server.

    But in cases where the block to be saved exceeds the maximum
    recommended request size (nbdClient.maxRequestSize), we
    need to split one big request into multiple not exceeding
    the limit

    If compression is enabled, function returns a list of
    offsets for the compressed frames, which is appended
    to the end of the stream.
    """
    wSize = 0
    cSizes = []
    for blocklen, blockOffset in block.step(
        blk.offset, blk.length, nbdCon.maxRequestSize
    ):
        if btype == "raw":
            writer.seek(blockOffset)

        try:
            data = nbdCon.nbd.pread(blocklen, blockOffset)
        except nbdError as e:
            raise DiskBackupFailed(e) from e

        if compress is not False:
            compressed = lz4.compressFrame(data, compress)
            wSize += writer.write(compressed)
            cSizes.append(len(compressed))
        else:
            wSize += writer.write(data)

        pbar.update(blocklen)

    return wSize, cSizes


def read(
    reader: IO[Any],
    offset: int,
    length: int,
    nbdCon,
    compression: bool,
    pbar,
) -> int:
    """Read data from reader and write to nbd connection

    If Compression is enabled function receives length information
    as dict, which contains the stream offsets for the compressed
    lz4 frames.

    Frames are read from the stream at the compressed size information
    (offset in the stream).

    After decompression, data is written back to original offset
    in the virtual machine disk image.

    If no compression is enabled, data is read from the regular
    data header at its position and written to nbd target
    directly.
    """
    wSize = 0
    for blocklen, blockOffset in block.step(offset, length, nbdCon.maxRequestSize):
        if compression is True:
            data = lz4.decompressFrame(reader.read(blocklen))
            nbdCon.nbd.pwrite(data, offset)
            offset += len(data)
            wSize += len(data)
        else:
            data = reader.read(blocklen)
            nbdCon.nbd.pwrite(data, blockOffset)
            wSize += len(data)

        pbar.update(blocklen)

    return wSize



#!/usr/bin/python3
"""
Copyright (C) 2023 Michael Ablassmeier <abi@grinser.de>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import logging
import pprint
from argparse import Namespace
from libvirtnbdbackup import chunk
from libvirtnbdbackup import lz4
from libvirtnbdbackup import common as lib
from libvirtnbdbackup.sparsestream import types
from libvirtnbdbackup.sparsestream import streamer
from libvirtnbdbackup.sparsestream.exceptions import StreamFormatException
from libvirtnbdbackup.exceptions import RestoreError
from libvirtnbdbackup.exceptions import UntilCheckpointReached


def restore(
    args: Namespace,
    stream: streamer.SparseStream,
    disk: str,
    targetFile: str,
    connection,
) -> bool:
    """Restore the data stream to the target file"""
    diskState = False
    diskState = _write(args, stream, disk, targetFile, connection)
    # no data has been processed
    if diskState is None:
        diskState = True

    return diskState


def _write(  # pylint: disable=too-many-branches,too-many-locals,too-many-statements
    args: Namespace,
    stream: streamer.SparseStream,
    dataFile: str,
    targetFile: str,
    connection,
) -> bool:
    """Restore data for disk"""
    sTypes = types.SparseStreamTypes()

    try:
        # pylint: disable=consider-using-with
        reader = open(dataFile, "rb")
    except OSError as errmsg:
        logging.error("Failed to open backup file for reading: [%s].", errmsg)
        raise RestoreError from errmsg

    try:
        kind, start, length = stream.readFrame(reader)
        meta = stream.loadMetadata(reader.read(length))
    except StreamFormatException as errmsg:
        logging.fatal(errmsg)
        raise RestoreError from errmsg

    trailer = None
    if lib.isCompressed(meta) is True:
        trailer = stream.readCompressionTrailer(reader)
        logging.info("Found compression trailer.")
        logging.debug("%s", trailer)

    if meta["dataSize"] == 0:
        logging.info("File [%s] contains no dirty blocks, skipping.", dataFile)
        if meta["checkpointName"] == args.until:
            logging.info("Reached checkpoint [%s], stopping", args.until)
            raise UntilCheckpointReached
        return True

    logging.info(
        "Applying data from backup file [%s] to target file [%s].", dataFile, targetFile
    )
    pprint.pprint(meta)
    assert reader.read(len(sTypes.TERM)) == sTypes.TERM

    progressBar = lib.progressBar(
        meta["dataSize"], f"restoring disk [{meta['diskName']}]", args
    )
    dataSize: int = 0
    dataBlockCnt: int = 0
    while True:
        try:
            kind, start, length = stream.readFrame(reader)
        except StreamFormatException as err:
            logging.error("Can't read stream at pos: [%s]: [%s]", reader.tell(), err)
            raise RestoreError from err
        if kind == sTypes.ZERO:
            logging.debug("Zero segment from [%s] length: [%s]", start, length)
        elif kind == sTypes.DATA:
            logging.debug(
                "Processing data segment from [%s] length: [%s]", start, length
            )

            originalSize = length
            if trailer:
                logging.debug("Block: [%s]", dataBlockCnt)
                logging.debug("Original block size: [%s]", length)
                length = trailer[dataBlockCnt]
                logging.debug("Compressed block size: [%s]", length)

            if originalSize >= connection.maxRequestSize:
                logging.debug(
                    "Chunked read/write, start: [%s], len: [%s]", start, length
                )
                try:
                    written = chunk.read(
                        reader,
                        start,
                        length,
                        connection,
                        lib.isCompressed(meta),
                        progressBar,
                    )
                except Exception as e:
                    logging.exception(e)
                    raise RestoreError from e
                logging.debug("Wrote: [%s]", written)
            else:
                try:
                    data = reader.read(length)
                    if lib.isCompressed(meta):
                        data = lz4.decompressFrame(data)
                    connection.nbd.pwrite(data, start)
                    written = len(data)
                except Exception as e:
                    logging.exception(e)
                    raise RestoreError from e
                progressBar.update(written)

            assert reader.read(len(sTypes.TERM)) == sTypes.TERM
            dataSize += originalSize
            dataBlockCnt += 1
        elif kind == sTypes.STOP:
            progressBar.close()
            if dataSize != meta["dataSize"]:
                logging.error(
                    "Restored data size does not match [%s] != [%s]",
                    dataSize,
                    meta["dataSize"],
                )
                raise RestoreError("Data size mismatch")
            break

    logging.info("End of stream, [%s] of data processed", lib.humanize(dataSize))
    if meta["checkpointName"] == args.until:
        logging.info("Reached checkpoint [%s], stopping", args.until)
        raise UntilCheckpointReached

    if connection.nbd.can_flush() is True:
        logging.debug("Flushing NBD connection handle")
        connection.nbd.flush()

    return True



"""
Copyright (C) 2023  Michael Ablassmeier <abi@grinser.de>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import logging
import socket
from typing import Tuple, Callable, Optional
from enum import Enum
from paramiko import (
    AutoAddPolicy,
    SSHClient,
    SFTPClient,
    SSHException,
    AuthenticationException,
)
from libvirtnbdbackup.ssh import exceptions
from libvirtnbdbackup.objects import processInfo

log = logging.getLogger("ssh")


class Mode(Enum):
    """Up or download mode"""

    UPLOAD = 1
    DOWNLOAD = 2


class client:
    """Wrapper around paramiko/sftp put and get functions, to be able to
    remote copy files from hypervisor host"""

    def __init__(
        self, host: str, user: str, port: int = 22, mode: Mode = Mode.DOWNLOAD
    ):
        self.client = None
        self.host = host
        self.user = user
        self.port = port
        self.copy: Callable[[str, str], None] = self.copyFrom
        if mode == Mode.UPLOAD:
            self.copy = self.copyTo
        self.connection = self.connect()
        self._sftp: Optional[SFTPClient] = None

    def connect(self) -> SSHClient:
        """Connect to remote system"""
        log.info(
            "Connecting remote system [%s] via ssh, username: [%s]",
            self.host,
            self.user,
        )
        try:
            cli = SSHClient()
            cli.load_system_host_keys()
            cli.set_missing_host_key_policy(AutoAddPolicy())
            cli.connect(
                self.host,
                username=self.user,
                port=self.port,
                timeout=5000,
            )
            return cli
        except AuthenticationException as e:
            raise exceptions.sshError(f"SSH key authentication failed: {e}")
        except socket.gaierror as e:
            raise exceptions.sshError(f"Unable to connect: {e}")
        except SSHException as e:
            raise exceptions.sshError(e)
        except Exception as e:
            log.exception(e)
            raise exceptions.sshError(f"Unhandled exception occurred: {e}")

    @property
    def sftp(self) -> SFTPClient:
        """Return SFTP client, opening connection on first use."""
        if self._sftp is None:
            self._sftp = self.connection.open_sftp()
        return self._sftp

    def exists(self, filepath: str) -> bool:
        """
        Check if remote file exists
        """
        try:
            self.sftp.stat(filepath)
            return True
        except IOError:
            return False

    def copyFrom(self, filepath: str, localpath: str) -> None:
        """
        Get file from remote system
        """
        log.info("Downloading file [%s] to [%s]", filepath, localpath)
        try:
            self.sftp.get(filepath, localpath)
        except SSHException as e:
            log.warning("Unable to download file: [%s]", e)

    def copyTo(self, localpath: str, remotepath: str) -> None:
        """
        Put file to remote system
        """
        log.info("Uploading file [%s] to [%s]", localpath, remotepath)
        try:
            self.sftp.put(localpath, remotepath)
        except SSHException as e:
            log.warning("Unable to upload file: [%s]", e)

    def _execute(self, cmd) -> Tuple[int, str, str]:
        _, stdout, stderr = self.connection.exec_command(cmd)
        ret = stdout.channel.recv_exit_status()
        err = stderr.read().strip().decode()
        out = stdout.read().strip().decode()
        return ret, err, out

    def run(self, cmd: str, pidFile: str = "", logFile: str = "") -> processInfo:
        """
        Execute command
        """
        pid: int = 0
        pidOut: str
        log.debug("Executing remote command: [%s]", cmd)
        ret, err, out = self._execute(cmd)
        logerr = ""
        if ret == 127:
            raise exceptions.sshError(err)
        if ret != 0:
            log.error(
                "Executing remote command failed, return code: [%s] stderr: [%s], stdout: [%s]",
                ret,
                err,
                out,
            )
            if logFile:
                log.debug("Attempting to catch errors from logfile: [%s]", logFile)
                _, _, logerr = self._execute(f"cat {logFile}")
            raise exceptions.sshError(
                f"Error during remote command: [{cmd}]: [{err}] [{logerr}]"
            )

        if pidFile:
            log.debug("PIDfile: [%s]", pidFile)
            _, _, pidOut = self._execute(f"cat {pidFile}")
            pid = int(pidOut)
            log.debug("PID: [%s]", pid)

        return processInfo(pid, logFile, err, out, pidFile)

    def disconnect(self):
        """Disconnect"""
        if self._sftp is not None:
            self._sftp.close()
            self._sftp = None
        if self.connection:
            self.connection.close()



#!/usr/bin/python3
"""
Copyright (C) 2023  Michael Ablassmeier <abi@grinser.de>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import os
import sys
from argparse import Namespace
from typing import Any
from libvirt import virDomain
from libvirtnbdbackup import virt
from libvirtnbdbackup import common as lib
from libvirtnbdbackup.objects import processInfo
from libvirtnbdbackup.qemu import util as qemu


class Backup:
    """Handle signal during backup operation"""

    @staticmethod
    def catch(
        args: Namespace,
        domObj: virDomain,
        virtClient: virt.client,
        log: Any,
        signum: int,
        _,
    ) -> None:
        """Catch signal, attempt to stop running backup job."""
        log.error("Signal caught: %s", signum)

        if args.offline is True:
            log.error("Exiting.")
            sys.exit(1)

        if virtClient.remoteHost != "":
            log.info("Reconnecting remote system to stop backup job.")
            try:
                virtClient = virt.client(args)
                domObj = virtClient.getDomain(args.domain)
            except virt.exceptions.connectionFailed as e:
                log.error("Reconnecting remote host failed: [%s]", e)
                log.error("Unable to stop backup job on remote system.", e)
                sys.exit(1)

        log.info("Cleanup: Stopping backup job.")
        virtClient.stopBackup(domObj)
        virtClient.close()
        sys.exit(1)


class Map:
    """Handle signal during map operation"""

    @staticmethod
    def catch(
        args: Namespace,
        nbdkitProcess: processInfo,
        blockMap,
        log: Any,
        signum,
        _,
    ):
        """Catch signal, attempt to stop processes."""
        log.info("Received signal: [%s]", signum)
        qemu.util("").disconnect(args.device)
        if not args.verbose:
            log.info("Removing temporary blockmap file: [%s]", blockMap.name)
            os.remove(blockMap.name)
            log.info("Removing nbdkit logfile: [%s]", nbdkitProcess.logFile)
            os.remove(nbdkitProcess.logFile)
        lib.killProc(nbdkitProcess.pid)
        sys.exit(0)



#!/usr/bin/python3
"""
Copyright (C) 2023 Michael Ablassmeier <abi@grinser.de>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import os
import zlib
import json
import logging
from typing import List
from argparse import Namespace
from libvirtnbdbackup import virt
from libvirtnbdbackup import output
from libvirtnbdbackup.restore import vmconfig
from libvirtnbdbackup.restore import header
from libvirtnbdbackup import common as lib
from libvirtnbdbackup.objects import DomainDisk
from libvirtnbdbackup.sparsestream import streamer
from libvirtnbdbackup.exceptions import RestoreError


def restore(args: Namespace, vmConfig: str, virtClient: virt.client) -> None:
    """Notice user if backed up vm had loader / nvram"""
    config = vmconfig.read(vmConfig)
    info = virtClient.getDomainInfo(config)

    for setting, val in info.items():
        f = lib.getLatest(args.input, f"*{os.path.basename(val)}*", -1)

        if len(f) == 0:
            logging.warning(
                "Unable to identify backup file for [%s] in specified backup folder [%s].",
                os.path.basename(val),
                args.input,
            )
            continue

        if lib.exists(args, val):
            logging.info(
                "File [%s]: for boot option [%s] already exists, skipping.",
                val,
                setting,
            )
            continue

        logging.info(
            "Restoring configured file [%s] for boot option [%s]", val, setting
        )
        lib.copy(args, f[0], val)


def verify(args: Namespace, dataFiles: List[str]) -> bool:
    """Compute adler32 checksum for exiting data files and
    compare with checksums computed during backup."""
    for dataFile in dataFiles:
        if args.disk is not None and not os.path.basename(dataFile).startswith(
            args.disk
        ):
            continue
        logging.debug("Using buffer size: %s", args.buffsize)
        logging.info("Computing checksum for: %s", dataFile)

        sourceFile = dataFile
        if args.sequence:
            sourceFile = os.path.join(args.input, dataFile)

        with output.openfile(sourceFile, "rb") as vfh:
            adler = 1
            data = vfh.read(args.buffsize)
            while data:
                adler = zlib.adler32(data, adler)
                data = vfh.read(args.buffsize)

        chksumFile = f"{sourceFile}.chksum"
        logging.info("Checksum result: %s", adler)
        if not os.path.exists(chksumFile):
            logging.info("No checksum found, skipping: [%s]", sourceFile)
            continue
        logging.info("Comparing checksum with stored information")
        with output.openfile(chksumFile, "r") as s:
            storedSum = int(s.read())
        if storedSum != adler:
            logging.error("Stored sums do not match: [%s]!=[%s]", storedSum, adler)
            return False

        logging.info("OK")
    return True


def dump(args: Namespace, stream: streamer.SparseStream, dataFiles: List[str]) -> bool:
    """Dump stream contents to json output"""
    logging.info("Dumping saveset meta information")
    entries = []
    for dataFile in dataFiles:
        if args.disk is not None and not os.path.basename(dataFile).startswith(
            args.disk
        ):
            continue
        logging.info(dataFile)

        sourceFile = dataFile
        if args.sequence:
            sourceFile = os.path.join(args.input, dataFile)

        try:
            meta = header.get(sourceFile, stream)
        except RestoreError as e:
            logging.error(e)
            continue

        entries.append(meta)

        if lib.isCompressed(meta):
            logging.info("Compressed stream found: [%s].", meta["compressionMethod"])

    print(json.dumps(entries, indent=4))

    return True


def target(args: Namespace, disk: DomainDisk) -> str:
    """Based on disk information, return target file
    to create during restore."""
    if disk.filename is not None:
        targetFile = os.path.join(args.output, disk.filename)
    else:
        targetFile = os.path.join(args.output, disk.target)

    return targetFile



"""
Copyright (C) 2023  Michael Ablassmeier <abi@grinser.de>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import logging
from lxml.etree import _Element

log = logging.getLogger()


def Optical(device: list, dev: str) -> bool:
    """Check if device is cdrom or floppy"""
    if device in ("cdrom", "floppy"):
        log.info("Skipping attached [%s] device: [%s].", device, dev)
        return True

    return False


def Lun(device: list, dev: str) -> bool:
    """Check if device is direct attached LUN"""
    if device == "lun":
        log.warning(
            "Skipping direct attached lun [%s], use option --raw to include.",
            dev,
        )
        return True

    return False


def Block(disk: _Element, dev: str) -> bool:
    """Check if device is direct attached block type device"""
    if disk.xpath("target")[0].get("type") == "block":
        log.warning(
            "Block device [%s] excluded by default, use option --raw to include.",
            dev,
        )
        return True

    return False


def Raw(diskFormat: str, dev: str) -> bool:
    """Check if disk has RAW disk format"""
    if diskFormat == "raw":
        log.warning(
            "Raw disk [%s] excluded by default, use option --raw to include.",
            dev,
        )
        return True

    return False



#!/usr/bin/python3
"""
Copyright (C) 2023 Michael Ablassmeier <abi@grinser.de>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import logging
from argparse import Namespace
from typing import Union
from libvirtnbdbackup import nbdcli
from libvirtnbdbackup import virt
from libvirtnbdbackup.qemu import util as qemu
from libvirtnbdbackup.ssh.exceptions import sshError
from libvirtnbdbackup.qemu.exceptions import ProcessError
from libvirtnbdbackup.exceptions import RestoreError

log = logging.getLogger("restore")


def setup(args: Namespace, exportName: str, targetFile: str, virtClient: virt.client):
    """Setup NBD process required for restore, either remote or local"""
    qFh = qemu.util(exportName)
    cType: Union[nbdcli.TCP, nbdcli.Unix]
    if not virtClient.remoteHost:
        logging.info("Starting local NBD server on socket: [%s]", args.socketfile)
        proc = qFh.startRestoreNbdServer(args, targetFile)
        cType = nbdcli.Unix(exportName, "", args.socketfile)
    else:
        remoteIP = virtClient.remoteHost
        if args.nbd_ip != "":
            remoteIP = args.nbd_ip
        logging.info(
            "Starting remote NBD server on socket: [%s:%s]",
            remoteIP,
            args.nbd_port,
        )
        proc = qFh.startRemoteRestoreNbdServer(args, targetFile)
        cType = nbdcli.TCP(exportName, "", remoteIP, args.tls, args.nbd_port)

    nbdClient = nbdcli.client(cType, False)
    logging.info("Started NBD server, PID: [%s]", proc.pid)
    return nbdClient.connect()


def start(args: Namespace, diskName: str, targetFile: str, virtClient: virt.client):
    """Start NDB Service"""
    try:
        return setup(args, diskName, targetFile, virtClient)
    except ProcessError as errmsg:
        logging.error(errmsg)
        raise RestoreError("Failed to start local NBD server.") from errmsg
    except sshError as errmsg:
        logging.error(errmsg)
        raise RestoreError("Failed to start remote NBD server.") from errmsg



#!/usr/bin/python3
"""
Copyright (C) 2023 Michael Ablassmeier <abi@grinser.de>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import logging
from argparse import Namespace
from libvirtnbdbackup import virt
from libvirtnbdbackup import common as lib
from libvirtnbdbackup.objects import DomainDisk
from libvirtnbdbackup.restore import server
from libvirtnbdbackup.restore import files
from libvirtnbdbackup.restore import image
from libvirtnbdbackup.restore import header
from libvirtnbdbackup.restore import data
from libvirtnbdbackup.restore import vmconfig
from libvirtnbdbackup.sparsestream import types
from libvirtnbdbackup.sparsestream import streamer
from libvirtnbdbackup.exceptions import RestoreError, UntilCheckpointReached
from libvirtnbdbackup.nbdcli.exceptions import NbdConnectionTimeout


def _backingstore(args: Namespace, disk: DomainDisk) -> None:
    """If an virtual machine was running on an snapshot image,
    warn user, the virtual machine configuration has to be
    adjusted before starting the VM is possible.

    User created external or internal Snapshots are not part of
    the backup.
    """
    if len(disk.backingstores) > 0 and not args.adjust_config:
        logging.warning(
            "Target image [%s] seems to be a snapshot image.", disk.filename
        )
        logging.warning("Target virtual machine configuration must be altered!")
        logging.warning("Configured backing store images must be changed.")


def restore(  # pylint: disable=too-many-branches,too-many-statements,too-many-locals
    args: Namespace, ConfigFile: str, virtClient: virt.client
) -> bytes:
    """Handle disk restore operation and adjust virtual machine
    configuration accordingly."""
    stream = streamer.SparseStream(types)
    vmConfig = vmconfig.read(ConfigFile)
    vmConfig = vmconfig.changeVolumePathes(args, vmConfig).decode()
    vmDisks = virtClient.getDomainDisks(args, vmConfig)
    if not vmDisks:
        raise RestoreError("Unable to parse disks from config")

    restConfig: bytes = vmConfig.encode()
    for disk in vmDisks:
        if args.disk not in (None, disk.target):
            logging.info("Skipping disk [%s] for restore", disk.target)
            continue

        restoreDisk = lib.getLatest(args.input, f"{disk.target}*.data")
        logging.debug("Restoring disk: [%s]", restoreDisk)
        if len(restoreDisk) < 1:
            logging.warning(
                "No backup file for disk [%s] found, assuming it has been excluded.",
                disk.target,
            )
            if args.adjust_config is True:
                restConfig = vmconfig.removeDisk(restConfig.decode(), disk.target)
            continue

        targetFile = files.target(args, disk)

        if args.raw and disk.format == "raw":
            logging.info("Restoring raw image to [%s]", targetFile)
            lib.copy(args, restoreDisk[0], targetFile)
            continue

        if "full" not in restoreDisk[0] and "copy" not in restoreDisk[0]:
            logging.error(
                "[%s]: Unable to locate base full or copy backup.", restoreDisk[0]
            )
            raise RestoreError("Failed to locate backup.")

        cptnum = -1
        if args.until is not None:
            cptnum = int(args.until.split(".")[-1])

        meta = header.get(restoreDisk[cptnum], stream)

        try:
            image.create(args, meta, targetFile, args.sshClient)
        except RestoreError as errmsg:
            raise RestoreError("Creating target image failed.") from errmsg

        try:
            connection = server.start(args, meta["diskName"], targetFile, virtClient)
        except NbdConnectionTimeout as e:
            raise RestoreError(e) from e

        for dataFile in restoreDisk:
            try:
                data.restore(args, stream, dataFile, targetFile, connection)
            except UntilCheckpointReached:
                break
            except RestoreError:
                break

        _backingstore(args, disk)
        if args.adjust_config is True:
            restConfig = vmconfig.adjust(args, disk, restConfig.decode(), targetFile)

        logging.debug("Closing NBD connection")
        connection.disconnect()

    if args.adjust_config is True:
        restConfig = vmconfig.removeUuid(restConfig.decode())
        restConfig = vmconfig.setVMName(args, restConfig.decode())

    return restConfig



"""
Exceptions
"""


class NbdClientException(Exception):
    """Nbd exceptions"""


class NbdConnectionError(NbdClientException):
    """Connection failed"""


class NbdConnectionTimeout(NbdClientException):
    """Connection timed out"""



#!/usr/bin/python3
"""
Copyright (C) 2023 Michael Ablassmeier <abi@grinser.de>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import os
from typing import List
from argparse import Namespace
from libvirtnbdbackup import virt
from libvirtnbdbackup import common as lib
from libvirtnbdbackup.restore import header
from libvirtnbdbackup.restore import server
from libvirtnbdbackup.restore import image
from libvirtnbdbackup.restore import data
from libvirtnbdbackup.sparsestream import types
from libvirtnbdbackup.sparsestream import streamer
from libvirtnbdbackup.exceptions import RestoreError


def restore(args: Namespace, dataFiles: List[str], virtClient: virt.client) -> bool:
    """Reconstruct image from a given set of data files"""
    stream = streamer.SparseStream(types)

    result: bool = False

    sourceFile = os.path.join(args.input, dataFiles[-1])
    meta = header.get(sourceFile, stream)
    if not meta:
        return result

    diskName = meta["diskName"]
    targetFile = os.path.join(args.output, diskName)
    if lib.exists(args, targetFile):
        raise RestoreError(f"Targetfile {targetFile} already exists.")

    try:
        image.create(args, meta, targetFile, args.sshClient)
    except RestoreError as errmsg:
        raise errmsg

    connection = server.start(args, diskName, targetFile, virtClient)

    for disk in dataFiles:
        sourceFile = os.path.join(args.input, disk)
        result = data.restore(args, stream, sourceFile, targetFile, connection)

    connection.disconnect()

    return result



#!/usr/bin/python3
"""
Copyright (C) 2023 Michael Ablassmeier <abi@grinser.de>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import logging
from typing import BinaryIO
from argparse import Namespace
from libvirtnbdbackup.virt.client import DomainDisk
from libvirtnbdbackup.common import getIdent, safeInfo


def get(
    args: Namespace, fileStream, targetFile: str, targetFilePartial: str
) -> BinaryIO:
    """Open target file based on output writer"""
    if args.stdout is True:
        logging.info("Writing data to zip archive.")
        fileStream.open(targetFile)
    else:
        safeInfo("Write data to target file: [%s].", targetFilePartial)
        fileStream.open(targetFilePartial)

    return fileStream


def Set(args: Namespace, disk: DomainDisk, ext: str = "data"):
    """Set Target file name to write data to, used for both data files
    and qemu disk info"""
    targetFile: str = ""
    level = args.level_filename
    if level in ("full", "copy"):
        if disk.format == "raw":
            level = "copy"
        targetFile = f"{args.output}/{disk.target}.{level}.{ext}"
    elif level in ("inc", "diff"):
        cptName = getIdent(args)
        targetFile = f"{args.output}/{disk.target}.{level}.{cptName}.{ext}"

    targetFilePartial = f"{targetFile}.partial"

    return targetFile, targetFilePartial



"""
Copyright (C) 2023  Michael Ablassmeier <abi@grinser.de>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import logging
from typing import List, Any, Generator, Dict
from nbd import CONTEXT_BASE_ALLOCATION
from libvirtnbdbackup.objects import Extent, _ExtentObj
from libvirtnbdbackup.common import humanize, safeInfo

log = logging.getLogger("extenthandler")


# pylint: disable=too-many-instance-attributes
class ExtentHandler:
    """Query extent information about allocated and
    zeroed regions from the NBD server started by libvirt/qemu

    This implementation should return the same
    extent information as nbdinfo or qemu-img map
    """

    def __init__(self, nbdFh, cType, no_sparse_detection: bool) -> None:
        self.useQemu = False
        self._maxRequestBlock: int = 4294967295
        self._align: int = 512
        self.lastExtentLen: int = 0
        self.offset: int = 0

        if nbdFh.__class__.__name__ == "util":
            self.useQemu = True
        self._nbdFh = nbdFh
        self._cType = cType
        self._extentEntries: Dict = {}
        self.no_sparse_detection = no_sparse_detection

        if cType.metaContext == "":
            self._metaContext = CONTEXT_BASE_ALLOCATION
        else:
            self._metaContext = cType.metaContext

        if self.useQemu is False:
            contexts = self._nbdFh.nbd.get_nr_meta_contexts()
            log.debug("NBD server exports [%d] metacontexts:", contexts)
            for i in range(0, contexts):
                ctx = self._nbdFh.nbd.get_meta_context(i)
                if self.no_sparse_detection is True and ctx == CONTEXT_BASE_ALLOCATION:
                    continue
                self._extentEntries[ctx] = []
        else:
            if self.no_sparse_detection is False:
                self._extentEntries[CONTEXT_BASE_ALLOCATION] = []
            self._extentEntries[self._metaContext] = []

        log.debug("Primary meta context for backup: %s", self._metaContext)

    def _getExtentCallback(
        self, metacontext: str, offset: int, entries: List, status: str
    ) -> None:
        """Callback function called by libnbd for each extent
        that is returned
        """
        log.debug("Metacontext is: %s", metacontext)
        log.debug("Offset is: %s", offset)
        log.debug("Status is: %s", status)
        self.lastExtentLen = len(self._extentEntries[self._metaContext])
        for entry in entries:
            self._extentEntries[metacontext].append(entry)
        log.debug("entries: %s", len(self._extentEntries[metacontext]))
        log.debug("Processed offsets: %s", self.offset)
        self.offset += sum(
            self._extentEntries[self._metaContext][self.lastExtentLen :: 2]
        )
        self.lastExtentLen = len(self._extentEntries[self._metaContext])

    def _setRequestAligment(self) -> int:
        """Align request size to nbd server"""
        align = self._nbdFh.nbd.get_block_size(0)
        if align == 0:
            align = self._align
        return self._maxRequestBlock - align + 1

    def queryExtents(self) -> List[Any]:
        """Query extents either via qemu or custom extent handler"""
        if self.useQemu:
            return self.queryExtentsQemu()

        return self.queryExtentsNbd()

    def queryExtentsQemu(self) -> List[Any]:
        """Use qemu utils to query extents from nbd server"""
        extents = []
        for ctx in iter(self._extentEntries):
            for extent in self._nbdFh.map(self._cType, ctx):
                extentObj = _ExtentObj(ctx, extent["length"], extent["type"])
                extents.append(extentObj)

        log.debug("Got %s extents from qemu command", len(extents))

        return extents

    def _extentsToObj(self) -> List[_ExtentObj]:
        """Go through extents and create a list of extent objects"""
        extentList = []
        for context, values in self._extentEntries.items():
            extentSizes = values[0::2]
            extentTypes = values[1::2]
            assert len(extentSizes) == len(extentTypes)
            ct = 0
            while ct < len(extentSizes):
                extentObj = _ExtentObj(context, extentSizes[ct], extentTypes[ct])
                extentList.append(extentObj)
                ct += 1

        return extentList

    @staticmethod
    def _unifyExtents(extentObjects: List[_ExtentObj]) -> Generator:
        """Unify extents. If a sequence of extents has the
        same type (data or zero) it is better to unify them
        into a bigger block, so during backup, less requests
        to the nbd server have to be sent
        """
        log.debug("Attempting to unify %s extents", len(extentObjects))
        cur = None
        for myExtent in extentObjects:
            if cur is None:
                cur = myExtent
            elif cur.type == myExtent.type and cur.context == myExtent.context:
                cur.length += myExtent.length
            else:
                yield cur
                cur = myExtent

        yield cur

    def queryExtentsNbd(self) -> List[_ExtentObj]:
        """Request used blocks/extents from the nbd service"""
        maxRequestLen = self._setRequestAligment()
        size = self._nbdFh.nbd.get_size()
        while self.offset < size:
            if size < maxRequestLen:
                request_length = size
            else:
                request_length = min(size - self.offset, maxRequestLen)
            log.debug("Block status request length: %s", request_length)
            self._nbdFh.nbd.block_status(
                request_length, self.offset, self._getExtentCallback
            )
            log.debug("Extents: %s", self._extentEntries)

        return self._extentsToObj()

    def setBlockType(self, context: str, blockType: int) -> bool:
        """Returns block type

        The extent types are as follows:

        For full backup:
            case 0  ("allocated")
            case 1: ("hole")
            case 2: ("zero")
            case 3: ("hole,zero")
        For checkpoint based inc/diff:
            case 0: ("clean")
            case 1: ("dirty")
        """
        data = None
        if context == CONTEXT_BASE_ALLOCATION:
            assert blockType in (0, 1, 2, 3)
            if blockType == 0:
                data = True
            if blockType == 1:
                data = False
            elif blockType == 2:
                data = True
            elif blockType == 3:
                data = False
        else:
            assert blockType in (0, 1)
            data = bool(blockType)

        assert data is not None
        return data

    def overlap(self, extents: List[Extent]) -> List[Extent]:
        """Find overlaps between base allocation and incremental bitmap to detect zero regions"""
        base_extents = [e for e in extents if e.context == CONTEXT_BASE_ALLOCATION]
        backup_extents = [
            e for e in extents if e.context == self._metaContext and e.data
        ]

        totalLength: int = 0
        result = []
        i = 0  # index for base_extents
        j = 0  # index for backup_extents
        while i < len(base_extents) and j < len(backup_extents):
            base = base_extents[i]
            backup = backup_extents[j]

            log.debug(
                "base: %d:%d(%s) bitmap: %d:%d(%s)",
                base.length,
                base.offset,
                str(base.data),
                backup.length,
                backup.offset,
                str(backup.data),
            )

            # Skip if either extent has data=False or no real intersection
            if not base.data or base.offset + base.length <= backup.offset:
                i += 1
                continue
            if not backup.data or backup.offset + backup.length <= base.offset:
                j += 1
                continue

            offset = max(base.offset, backup.offset)
            end = min(backup.offset + backup.length, base.offset + base.length)
            log.debug("-->: %d:%d", offset, end - offset)
            result.append(
                Extent(
                    context=base.context,
                    data=True,
                    offset=offset,
                    length=end - offset,
                )
            )
            totalLength += end - offset

            # advance
            if end == base.offset + base.length:
                i += 1
            if end == backup.offset + backup.length:
                j += 1

        if totalLength > 0:
            safeInfo(
                "Detected %d bytes [%s] non-sparse blocks for current bitmap.",
                totalLength,
                humanize(totalLength),
            )

        return result

    def queryBlockStatus(self) -> List[Extent]:
        """Check the status for each extent, whether if it is
        real data or zeroes, return a list of extent objects
        """
        safeInfo("Start receiving backup extents.")
        extents = self.queryExtents()
        safeInfo("Finished receiving extents.")
        extentList: List[Extent] = []
        start: int = 0
        baseStart: int = 0
        totalLength: int = 0
        for extent in self._unifyExtents(extents):
            extObj = Extent(
                extent.context,
                self.setBlockType(extent.context, extent.type),
                baseStart if extent.context == CONTEXT_BASE_ALLOCATION else start,
                extent.length,
            )
            extentList.append(extObj)
            if extent.context == CONTEXT_BASE_ALLOCATION:
                baseStart += extent.length
            else:
                start += extent.length
                if extObj.data:
                    totalLength += extent.length
            log.debug(
                "%s %d %d %d",
                extObj.context,
                extObj.data,
                extObj.offset,
                extObj.offset + extObj.length,
            )
        if self.no_sparse_detection is True:
            safeInfo("Skipping detection of sparse/fstrimmed blocks.")
            return extentList

        if self._metaContext != CONTEXT_BASE_ALLOCATION:
            log.debug("Detected [%d] bytes of changed data regions.", totalLength)
            extentList = self.overlap(extentList)
        return extentList



"""
Copyright (C) 2023  Michael Ablassmeier <abi@grinser.de>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from argparse import Namespace
from typing import Union
from libvirtnbdbackup.output.target.directory import Directory
from libvirtnbdbackup.output.target.zip import Zip


def get(
    args: Namespace,
) -> Union[Directory, Zip]:
    """Get filehandle for output files based on output
    mode"""
    fileStream: Union[Directory, Zip]
    if args.stdout is False:
        fileStream = Directory()
    else:
        fileStream = Zip()
        args.output = "./"
        args.worker = 1

    return fileStream



"""
Exceptions
"""


class QemuHelperError(Exception):
    """Errors during qemu helper"""


class NbdServerProcessError(QemuHelperError):
    """Unable to start nbd server for offline backup"""


class ProcessError(QemuHelperError):
    """Unable to start process"""



#!/usr/bin/python3
"""
Copyright (C) 2023 Michael Ablassmeier <abi@grinser.de>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import os
import logging
import json
from argparse import Namespace
from typing import List, Dict
from libvirtnbdbackup.qemu import util as qemu
from libvirtnbdbackup import output
from libvirtnbdbackup import common as lib
from libvirtnbdbackup.exceptions import RestoreError
from libvirtnbdbackup.qemu.exceptions import ProcessError
from libvirtnbdbackup.output.exceptions import OutputException
from libvirtnbdbackup.ssh.exceptions import sshError


def getConfig(  # pylint: disable=too-many-statements
    args: Namespace, meta: Dict[str, str]
) -> List[str]:
    """Check if backup includes exported qcow config and return a list
    of options passed to qemu-img create command"""
    opt: List[str] = []
    qcowConfig = None
    qcowConfigFile = lib.getLatest(args.input, f"{meta['diskName']}*.qcow.json*", -1)
    if not qcowConfigFile:
        logging.warning(
            "No QCOW image config found in [%s], will use default options.", args.input
        )
        return opt

    lastConfigFile = qcowConfigFile[0]

    try:
        with output.openfile(lastConfigFile, "rb") as qFh:
            qcowConfig = json.loads(qFh.read().decode())
        logging.info("Using QCOW options from backup file: [%s]", lastConfigFile)
    except (
        OutputException,
        json.decoder.JSONDecodeError,
    ) as errmsg:
        logging.warning(
            "Unable to load original QCOW image config, using defaults: [%s].",
            errmsg,
        )
        return opt

    try:
        opt.append("-o")
        opt.append(f"compat={qcowConfig['format-specific']['data']['compat']}")
    except KeyError as errmsg:
        logging.warning("Unable apply QCOW specific compat option: [%s]", errmsg)

    try:
        opt.append("-o")
        opt.append(f"cluster_size={qcowConfig['cluster-size']}")
    except KeyError as errmsg:
        logging.warning("Unable apply QCOW specific cluster_size option: [%s]", errmsg)

    try:
        if qcowConfig["format-specific"]["data"]["lazy-refcounts"]:
            opt.append("-o")
            opt.append("lazy_refcounts=on")
    except KeyError as errmsg:
        logging.warning(
            "Unable apply QCOW specific lazy_refcounts option: [%s]", errmsg
        )

    try:
        cType = qcowConfig["format-specific"]["data"]["compression-type"]
        opt.append("-o")
        opt.append(f"compression_type={cType}")
        logging.info("Setting image compression type: [%s]", cType)
    except KeyError as errmsg:
        pass

    try:
        dataFile = qcowConfig["format-specific"]["data"]["data-file"]
        if args.adjust_config is True:
            dataFilePath = os.path.join(
                args.output,
                os.path.basename(dataFile),
            )
            logging.info(
                "QCOW image with data-file backend detected: [%s], adjusting path to: [%s]",
                dataFile,
                dataFilePath,
            )
        else:
            logging.info(
                "QCOW image with data-file backend detected, keeping original path: [%s]",
                dataFile,
            )

        opt.append("-o")
        opt.append(f"data_file={dataFilePath}")
    except KeyError as errmsg:
        pass

    try:
        if qcowConfig["format-specific"]["data"]["data-file-raw"] is True:
            opt.append("-o")
            opt.append("data_file_raw=true")
        logging.info("QCOW image with RAW data-file backend detected.")
    except KeyError as errmsg:
        pass

    return opt


def create(args: Namespace, meta: Dict[str, str], targetFile: str, sshClient):
    """Read QCOW image related backup json and create target image file using
    its original options"""
    options = getConfig(args, meta)
    logging.info(
        "Create virtual disk [%s] format: [%s] size: [%s] based on: [%s] preallocated: [%s]",
        targetFile,
        meta["diskFormat"],
        meta["virtualSize"],
        meta["checkpointName"],
        args.preallocate,
    )

    if lib.exists(args, targetFile):
        logging.error(
            "Target file already exists: [%s], won't overwrite.",
            os.path.abspath(targetFile),
        )
        raise RestoreError

    qFh = qemu.util(meta["diskName"])
    try:
        qFh.create(
            args,
            targetFile,
            int(meta["virtualSize"]),
            meta["diskFormat"],
            options,
            sshClient,
        )
    except (ProcessError, sshError) as e:
        logging.error("Failed to create restore target: [%s]", e)
        raise RestoreError from e



"""
Copyright (C) 2023  Michael Ablassmeier <abi@grinser.de>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from typing import Generator, IO, Any, Union
from nbd import Error as nbdError
from libvirtnbdbackup import lz4
from libvirtnbdbackup.exceptions import BackupException


def step(offset: int, length: int, maxRequestSize: int) -> Generator:
    """Process block and ensure to not exceed the maximum request size
    from NBD server.

    If length parameter is dict, compression was enabled during
    backup, thus we cannot use the offsets and sizes for the
    original data, but must use the compressed offsets and sizes
    to read the correct lz4 frames from the stream.
    """
    blockOffset = offset
    if isinstance(length, dict):
        blockOffset = offset
        compressOffset = list(length.keys())[0]
        for part in length[compressOffset]:
            blockOffset += part
            yield part, blockOffset
    else:
        blockOffset = offset
        while blockOffset < offset + length:
            blocklen = min(offset + length - blockOffset, maxRequestSize)
            yield blocklen, blockOffset
            blockOffset += blocklen


def write(
    writer: IO[Any], block, nbdCon, btype: str, compress: Union[bool, int]
) -> int:
    """Write single block that does not exceed nbd maxRequestSize
    setting. In case compression is enabled, single blocks are
    compressed using lz4.
    """
    if btype == "raw":
        writer.seek(block.offset)

    try:
        data = nbdCon.nbd.pread(block.length, block.offset)
    except nbdError as e:
        raise BackupException(e) from e

    if compress is not False:
        data = lz4.compressFrame(data, compress)

    return writer.write(data)



"""
Copyright (C) 2023  Michael Ablassmeier <abi@grinser.de>
Copyright (C) 2020 Red Hat, Inc.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import json
import os
import datetime
from typing import List, Any, Tuple, Dict
from argparse import Namespace
from libvirtnbdbackup.objects import DomainDisk
from libvirtnbdbackup.sparsestream import exceptions


class SparseStream:
    """Sparse Stream writer/reader class"""

    def __init__(self, types, version: int = 2) -> None:
        """Stream version:

        1: base version
        2: stream version with compression support
        """
        self.version = version
        self.compressionMethod: str = "lz4"
        self.types = types.SparseStreamTypes()

    def dumpMetadata(
        self,
        args: Namespace,
        virtualSize: int,
        dataSize: int,
        disk: DomainDisk,
    ) -> bytes:
        """First block in backup stream is Meta data information
        about virtual size of the disk being backed up, as well
        as various information regarding backup.
        Dumps Metadata frame to be written at start of stream in
        json format.
        """
        meta = {
            "virtualSize": virtualSize,
            "dataSize": dataSize,
            "date": datetime.datetime.now().isoformat(),
            "diskName": disk.target,
            "diskFormat": disk.format,
            "checkpointName": args.cpt.name,
            "compressed": args.compress,
            "compressionMethod": self.compressionMethod,
            "parentCheckpoint": args.cpt.parent,
            "incremental": (args.level in ("inc", "diff")),
            "streamVersion": self.version,
        }
        return json.dumps(meta, indent=4).encode("utf-8")

    def writeCompressionTrailer(self, writer, trailer: List[Any]) -> None:
        """Dump compression trailer to end of stream"""
        size = writer.write(json.dumps(trailer).encode())
        writer.write(self.types.TERM)
        self.writeFrame(writer, self.types.COMP, 0, size)

    def _readHeader(self, reader) -> Tuple[str, str, str]:
        """Attempt to read header"""
        header = reader.read(self.types.FRAME_LEN)
        try:
            kind, start, length = header.split(b" ", 2)
        except ValueError as err:
            raise exceptions.BlockFormatException(
                f"Invalid block format: [{err}]"
            ) from err

        return kind, start, length

    @staticmethod
    def _parseHeader(kind, start: str, length: str) -> Tuple[str, int, int]:
        """Return parsed header information"""
        try:
            return kind, int(start, 16), int(length, 16)
        except ValueError as err:
            raise exceptions.FrameformatException(
                f"Invalid frame format: [{err}]"
            ) from err

    def readCompressionTrailer(self, reader) -> Dict[int, Any]:
        """If compressed stream is found, information about compressed
        block sizes is appended as last json payload.

        Function seeks to end of file and reads trailer information.
        """
        pos = reader.tell()
        reader.seek(0, os.SEEK_END)
        reader.seek(-(self.types.FRAME_LEN + len(self.types.TERM)), os.SEEK_CUR)
        _, _, length = self._readHeader(reader)
        reader.seek(-(self.types.FRAME_LEN + int(length, 16)), os.SEEK_CUR)
        trailer = self.loadMetadata(reader.read(int(length, 16)))
        reader.seek(pos)
        return trailer

    @staticmethod
    def loadMetadata(s: bytes) -> Any:
        """Load and parse metadata information
        Parameters:
            s:  (str)   Json string as received during data file read
        Returns:
            json.loads: (dict)  Decoded json string as python object
        """
        try:
            return json.loads(s.decode("utf-8"))
        except json.decoder.JSONDecodeError as err:
            raise exceptions.MetaHeaderFormatException(
                f"Invalid meta header format: [{err}]"
            ) from err

    def writeFrame(self, writer, kind, start: int, length: int) -> None:
        """Write backup frame
        Parameters:
            writer: (fh)    Writer object that implements .write()
        """
        writer.write(self.types.FRAME % (kind, start, length))

    def readFrame(self, reader) -> Tuple[str, int, int]:
        """Read backup frame
        Parameters:
            reader: (fh)    Reader object which implements .read()
        """
        kind, start, length = self._readHeader(reader)
        return self._parseHeader(kind, start, length)



"""
Copyright (C) 2023  Michael Ablassmeier <abi@grinser.de>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import logging
import lz4.frame

log = logging.getLogger()


def decompressFrame(data: bytes) -> bytes:
    """Decompress lz4 frame, print frame information"""
    frameInfo = lz4.frame.get_frame_info(data)
    log.debug("Compressed Frame: %s", frameInfo)
    return lz4.frame.decompress(data)


def compressFrame(data: bytes, level: int) -> bytes:
    """Compress block with to lz4 frame, checksums
    enabled for safety
    """
    return lz4.frame.compress(
        data,
        content_checksum=True,
        block_checksum=True,
        compression_level=level,
    )



"""
Copyright (C) 2023  Michael Ablassmeier <abi@grinser.de>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import os
import logging
from time import sleep
import nbd
from libvirtnbdbackup.nbdcli import exceptions

log = logging.getLogger("nbd")


# pylint: disable=too-many-instance-attributes
class client:
    """Helper functions for NBD"""

    def __init__(self, cType, no_sparse_detection: bool):
        """
        Connect NBD backend
        """
        self.cType = cType
        self._exportName = cType.exportName
        self._metaContext = ""
        if cType.metaContext != "":
            self._metaContext = cType.metaContext
        else:
            self._metaContext = nbd.CONTEXT_BASE_ALLOCATION
        self.maxRequestSize = 33554432
        self.minRequestSize = 65536
        self.no_sparse_detection = no_sparse_detection
        self.nbd = nbd.NBD()

        def debug(func, args):
            """Write NBD debugging messages to logfile instead of
            stderr"""
            log.debug("%s: %s", func, args)

        self.nbd.set_debug_callback(debug)
        self.connection = None

    def _getBlockInfo(self) -> None:
        """Read maximum request/block size as advertised by the nbd
        server. This is the value which will then be used by default
        """
        maxSize = self.nbd.get_block_size(nbd.SIZE_MAXIMUM)
        if maxSize != 0:
            self.maxRequestSize = maxSize

        log.debug("Block size supported by NBD server: [%s]", maxSize)

    def _connect(self) -> nbd.NBD:
        """Setup connection to NBD server endpoint, return
        connection handle
        """
        if self.cType.tls and not self.nbd.supports_tls():
            raise exceptions.NbdConnectionError(
                "Installed python nbd binding is missing required tls features."
            )

        try:
            if self.cType.tls:
                self.nbd.set_tls(nbd.TLS_REQUIRE)
            if self.no_sparse_detection is False:
                self.nbd.add_meta_context(nbd.CONTEXT_BASE_ALLOCATION)
            if self._metaContext != "":
                log.debug(
                    "Adding meta context to NBD connection: [%s]", self._metaContext
                )
                self.nbd.add_meta_context(self._metaContext)
            self.nbd.set_export_name(self._exportName)
            self.nbd.connect_uri(self.cType.uri)
        except nbd.Error as e:
            raise exceptions.NbdConnectionError(f"Unable to connect nbd server: {e}")

        self._getBlockInfo()

        return self.nbd

    def connect(self) -> nbd.NBD:
        """Wait until NBD endpoint connection can be established. It can take
        some time until qemu-nbd process is running and reachable. Attempt to
        connect and fail if no connection can be established. In case of unix
        domain socket, wait until socket file is created by qemu-nbd."""
        log.info("Waiting until NBD server at [%s] is up.", self.cType.uri)
        for retry in range(20):
            sleep(1)
            if self.cType.backupSocket and not os.path.exists(self.cType.backupSocket):
                log.info("Waiting for NBD Server unix socket, Retry: %s", retry)
                continue

            try:
                connection = self._connect()
            except exceptions.NbdConnectionError as e:
                self.nbd = nbd.NBD()
                log.info("Waiting for NBD Server connection, Retry: %s [%s]", retry, e)
                continue

            log.info("Connection to NBD backend succeeded.")
            self.connection = connection
            return self

        raise exceptions.NbdConnectionTimeout(
            "Timeout during connection to NBD server backend."
        )

    def disconnect(self) -> None:
        """Close nbd connection handle"""
        self.nbd.shutdown()



#!/usr/bin/python3
"""
Copyright (C) 2023 Michael Ablassmeier <abi@grinser.de>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import logging
from argparse import Namespace
from typing import List, Any, Tuple
from libvirtnbdbackup import nbdcli
from libvirtnbdbackup import virt
from libvirtnbdbackup.virt.client import DomainDisk
from libvirtnbdbackup.objects import processInfo
from libvirtnbdbackup.sparsestream import streamer
from libvirtnbdbackup.sparsestream import types
from libvirtnbdbackup import exceptions
from libvirtnbdbackup import chunk
from libvirtnbdbackup import block
from libvirtnbdbackup.backup import partialfile
from libvirtnbdbackup.backup import server
from libvirtnbdbackup.backup import target
from libvirtnbdbackup.backup.metadata import backupChecksum
from libvirtnbdbackup import extenthandler
from libvirtnbdbackup.qemu import util as qemu
from libvirtnbdbackup.qemu.exceptions import ProcessError
from libvirtnbdbackup import common as lib
from libvirtnbdbackup.output import stream


def _setStreamType(args: Namespace, disk: DomainDisk) -> str:
    """Set target stream type based on disk format"""
    streamType = "raw"
    if disk.format != streamType:
        streamType = args.type

    return streamType


def _getExtentHandler(args: Namespace, nbdClient):
    """Query dirty blocks either via qemu client or self
    implemented extend handler"""
    if args.qemu:
        lib.safeInfo("Using qemu tools to query extents")
        extentHandler = extenthandler.ExtentHandler(
            qemu.util(nbdClient.cType.exportName),
            nbdClient.cType,
            args.no_sparse_detection,
        )
    else:
        extentHandler = extenthandler.ExtentHandler(
            nbdClient, nbdClient.cType, args.no_sparse_detection
        )

    return extentHandler


def backup(  # pylint: disable=too-many-arguments,too-many-branches, too-many-locals, too-many-statements
    args: Namespace,
    disk: DomainDisk,
    count: int,
    fileStream,
    virtClient: virt.client,
) -> Tuple[int, bool]:
    """Backup domain disk data."""
    dStream = streamer.SparseStream(types)
    sTypes = types.SparseStreamTypes()
    lib.setThreadName(disk.target)
    streamType = _setStreamType(args, disk)
    metaContext = nbdcli.context.get(args, disk)
    nbdProc: processInfo
    remoteIP: str = virtClient.remoteHost
    port: int = args.nbd_port
    if args.nbd_ip != "":
        remoteIP = args.nbd_ip

    if args.offline is True:
        port = args.nbd_port + count
        try:
            nbdProc = server.setup(args, disk, virtClient.remoteHost, port)
        except ProcessError as errmsg:
            logging.error(errmsg)
            raise exceptions.DiskBackupFailed("Failed to start NBD server.")

    if disk.discardOption is not None:
        lib.safeInfo("Virtual disk discard option: [%s]", disk.discardOption)

    connection = server.connect(args, disk, metaContext, remoteIP, port, virtClient)

    extentHandler = _getExtentHandler(args, connection)
    extents = extentHandler.queryBlockStatus()
    diskSize = connection.nbd.get_size()

    if extents is None:
        logging.error("No extents returned by NBD server.")
        return 0, False

    thinBackupSize = sum(extent.length for extent in extents if extent.data is True)
    lib.safeInfo("Got %s extents to backup.", len(extents))
    logging.debug("%s", lib.dumpExtentJson(extents))
    lib.safeInfo("%s bytes [%s] virtual disk size", diskSize, lib.humanize(diskSize))
    lib.safeInfo(
        "%s bytes [%s] of data extents to backup",
        thinBackupSize,
        lib.humanize(thinBackupSize),
    )

    if args.level in ("inc", "diff") and thinBackupSize == 0:
        lib.safeInfo("No dirty blocks found")
        args.noprogress = True

    targetFile, targetFilePartial = target.Set(args, disk)

    # if writing to regular files we want instantiate an new
    # handle for each file otherwise multiple threads collid
    # during file close
    # in case of zip file output we want to use the existing
    # opened output channel
    if not args.stdout:
        fileStream = stream.get(args)
    writer = target.get(args, fileStream, targetFile, targetFilePartial)

    if streamType == "raw":
        lib.safeInfo("Creating full provisioned raw backup image")
        writer.truncate(diskSize)
    else:
        lib.safeInfo("Creating thin provisioned stream backup image")
        header = dStream.dumpMetadata(
            args,
            diskSize,
            thinBackupSize,
            disk,
        )
        dStream.writeFrame(writer, sTypes.META, 0, len(header))
        writer.write(header)
        writer.write(sTypes.TERM)

    progressBar = lib.progressBar(
        thinBackupSize, f"saving disk {disk.target}", args, count=count
    )
    compressedSizes: List[Any] = []
    backupSize: int = 0
    for save in extents:
        if save.data is True:
            if streamType == "stream":
                dStream.writeFrame(writer, sTypes.DATA, save.offset, save.length)
                logging.debug(
                    "Read data from: start %s, length: %s", save.offset, save.length
                )

            cSizes = None

            if save.length >= connection.maxRequestSize:
                logging.debug(
                    "Chunked data read from: start %s, length: %s",
                    save.offset,
                    save.length,
                )
                size, cSizes = chunk.write(
                    writer, save, connection, streamType, args.compress, progressBar
                )
            else:
                size = block.write(
                    writer,
                    save,
                    connection,
                    streamType,
                    args.compress,
                )
                if streamType == "raw":
                    size = writer.seek(save.offset)

                progressBar.update(save.length)

            if streamType == "stream":
                writer.write(sTypes.TERM)
                if args.compress:
                    logging.debug("Compressed size: %s", size)
                    backupSize += size
                    if cSizes:
                        blockList = {}
                        blockList[size] = cSizes
                        compressedSizes.append(blockList)
                    else:
                        compressedSizes.append(size)
                else:
                    assert size == save.length
                    backupSize += save.length
        else:
            if streamType == "raw":
                writer.seek(save.offset)
                backupSize += save.length
            elif streamType == "stream" and args.level not in ("inc", "diff"):
                dStream.writeFrame(writer, sTypes.ZERO, save.offset, save.length)
    if streamType == "stream":
        dStream.writeFrame(writer, sTypes.STOP, 0, 0)
        if args.compress:
            dStream.writeCompressionTrailer(writer, compressedSizes)

    progressBar.close()
    writer.close()
    connection.disconnect()

    if args.offline is True and virtClient.remoteHost == "":
        lib.safeInfo("Stopping NBD Service.")
        lib.killProc(nbdProc.pid)

    if args.offline is True:
        lib.remove(args, nbdProc.pidFile)

    if not args.stdout:
        if args.noprogress is True:
            lib.safeInfo(
                "Backup of disk [%s] finished, file: [%s]", disk.target, targetFile
            )
        partialfile.rename(targetFilePartial, targetFile)
    if streamType != "raw":
        backupChecksum(fileStream, targetFile)

    return backupSize, True



"""
Copyright (C) 2023  Michael Ablassmeier <abi@grinser.de>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import logging
from argparse import Namespace
from libvirtnbdbackup.virt.client import DomainDisk

log = logging.getLogger("nbdctx")


def get(args: Namespace, disk: DomainDisk) -> str:
    """Get required meta context string passed to nbd server based on
    backup type"""
    metaContext = ""
    if args.level not in ("inc", "diff"):
        return metaContext

    if args.offline is True:
        metaContext = f"qemu:dirty-bitmap:{args.cpt.name}"
    else:
        metaContext = f"qemu:dirty-bitmap:backup-{disk.target}"

    logging.debug("Using NBD meta context [%s]", metaContext)

    return metaContext



"""
Copyright (C) 2024  Michael Ablassmeier <abi@grinser.de>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import logging
from typing import List, Any
from argparse import Namespace
from libvirt import virDomain
from libvirtnbdbackup import virt
from libvirtnbdbackup import common as lib
from libvirtnbdbackup import exceptions

log = logging.getLogger()


def arguments(args: Namespace) -> None:
    """Check passed arguments for validity"""
    if args.compress is not False and args.type == "raw":
        raise exceptions.BackupException("Compression not supported with raw output.")

    if args.stdout is True and args.type == "raw":
        raise exceptions.BackupException("Output type raw not supported to stdout.")

    if args.stdout is True and args.raw is True:
        raise exceptions.BackupException(
            "Saving raw images to stdout is not supported."
        )

    if args.type == "raw" and args.level in ("inc", "diff"):
        raise exceptions.BackupException(
            "Stream format raw does not support incremental or differential backup."
        )


def targetDir(args: Namespace) -> None:
    """Check if target directory backup is started to meets
    all requirements based on the backup level executed"""
    if (
        args.level not in ("copy", "full", "auto")
        and not lib.hasFullBackup(args)
        and not args.stdout
    ):
        raise exceptions.BackupException(
            f"Unable to execute [{args.level}] backup: "
            f"No full backup found in target directory: [{args.output}]"
        )

    if lib.targetIsEmpty(args) and args.level == "auto":
        log.info("Backup mode auto, target folder is empty: executing full backup.")
        args.level = "full"
    elif not lib.targetIsEmpty(args) and args.level == "auto":
        if not lib.hasFullBackup(args):
            raise exceptions.BackupException(
                "Can't execute switch to auto incremental backup: "
                f"specified target folder [{args.output}] does not contain full backup.",
            )
        log.info("Backup mode auto: executing incremental backup.")
        args.level = "inc"
    elif not args.stdout and not args.startonly and not args.killonly:
        if not lib.targetIsEmpty(args):
            raise exceptions.BackupException(
                "Target directory already contains full or copy backup."
            )


def vmstate(args, virtClient: virt.client, domObj: virDomain) -> None:
    """Check virtual machine state before executing backup
    and based on situation, either fallback to regular copy
    backup or attempt to bring VM into paused state"""
    if domObj.isActive() == 0:
        args.offline = True
        if args.start_domain is True:
            log.info("Starting domain in paused state")
            if virtClient.startDomain(domObj) == 0:
                args.offline = False
            else:
                log.info("Failed to start VM in paused mode.")

    if args.level == "full" and args.offline is True:
        log.warning("Domain is offline, resetting backup options.")
        args.level = "copy"
        log.warning("New Backup level: [%s].", args.level)

    if args.offline is True and args.startonly is True:
        raise exceptions.BackupException(
            "Domain is offline: must be active for this function."
        )


def vmfeature(virtClient: virt.client, domObj: virDomain) -> None:
    """Check if required features are enabled in domain config"""
    if virtClient.hasIncrementalEnabled(domObj) is False:
        raise exceptions.BackupException(
            "Virtual machine does not support required backup features, "
            "please adjust virtual machine configuration."
        )


def diskformat(args: Namespace, disks: List[Any]) -> None:
    """Check if disks meet requirements for backup mode, if not, switch
    backup job to type copy."""
    if args.level != "copy" and lib.hasQcowDisks(disks) is False:
        args.level = "copy"
        raise exceptions.BackupException(
            "Only raw disks attached, switching to backup mode copy."
        )


def blockjobs(
    args, virtClient: virt.client, domObj: virDomain, disks: List[Any]
) -> None:
    """Check if there is an already active backup operation on the domain
    disks. If so, fail accordingly"""
    if (
        not args.killonly
        and not args.offline
        and virtClient.blockJobActive(domObj, disks)
    ):
        raise exceptions.BackupException(
            "Active block job for running domain:"
            f" Check with [virsh domjobinfo {args.domain}] or use option -k to kill the active job."
        )



"""
Copyright (C) 2023  Michael Ablassmeier <abi@grinser.de>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import ipaddress
from dataclasses import dataclass


@dataclass
class processInfo:
    """Process info object returned by functions calling
    various qemu commands
    """

    pid: int
    logFile: str
    err: str
    out: str
    pidFile: str


@dataclass
class DomainDisk:
    """Domain disk object holding information about the disk
    attached to a virtual machine"""

    target: str
    format: str
    filename: str
    path: str
    backingstores: list
    discardOption: str


@dataclass
class nbdConn:
    """NBD connection"""

    exportName: str
    metaContext: str


@dataclass
class Unix(nbdConn):
    """NBD connection type unix for connection via socket file"""

    backupSocket: str
    tls: bool = False

    def __post_init__(self):
        self.uri = f"nbd+unix:///{self.exportName}?socket={self.backupSocket}"


@dataclass
class TCP(nbdConn):
    """NBD connection type tcp for remote backup"""

    hostname: str
    tls: bool
    port: int = 10809
    backupSocket: str = ""
    uri_prefix = "nbd://"

    def __post_init__(self):
        if self.tls:
            self.uri_prefix = "nbds://"

        try:
            ip = ipaddress.ip_address(self.hostname)
            if ip.version == 6:
                self.hostname = f"[{self.hostname}]"
        except ValueError:
            pass

        self.uri = f"{self.uri_prefix}{self.hostname}:{self.port}/{self.exportName}"


@dataclass
class Extent:
    """Extent description containing information if block contains
    data, offset and length of data to be read/written"""

    context: str
    data: bool
    offset: int
    length: int


@dataclass
class _ExtentObj:
    """Single Extent object as returned from the NBD server"""

    context: str
    length: int
    type: int



#!/usr/bin/python3
"""
Copyright (C) 2023  Michael Ablassmeier <abi@grinser.de>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import os
import sys
import shutil
import logging
from argparse import Namespace
from libvirtnbdbackup import common as lib


def executables() -> None:
    """Check if required utils are installed"""
    for exe in ("nbdkit", "qemu-nbd"):
        if not shutil.which(exe):
            logging.error("Please install required [%s] utility.", exe)


def device(args: Namespace) -> None:
    """Check if /dev/nbdX exists, otherwise it is likely
    nbd module isn't loaded on the system"""
    if not args.device.startswith("/dev/nbd"):
        logging.error("Target device [%s] seems not to be an NBD device?", args.device)

    if not lib.exists(args, args.device):
        logging.error(
            "Target device [%s] does not exist, please load nbd module: [modprobe nbd]",
            args.device,
        )


def plugin(args: Namespace) -> str:
    """Attempt to locate the nbdkit plugin that is passed to the
    nbdkit process"""
    pluginFileName = "virtnbd-nbdkit-plugin"
    installDir = os.path.dirname(sys.argv[0])
    nbdkitModule = f"{installDir}/{pluginFileName}"

    if not lib.exists(args, nbdkitModule):
        logging.error("Failed to locate nbdkit plugin: [%s]", pluginFileName)

    return nbdkitModule



"""
Copyright (C) 2023  Michael Ablassmeier <abi@grinser.de>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import os
import sys
import zlib
import logging
import builtins
from typing import IO, Union, Any
from libvirtnbdbackup.output import exceptions

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal

log = logging.getLogger("directory")


class Directory:
    """Backup to target directory"""

    def __init__(self) -> None:
        self.fileHandle: IO[Any]
        self.chksum: int = 1

    def create(self, targetDir) -> None:
        """Create wrapper"""
        log.debug("Create: %s", targetDir)
        if os.path.exists(targetDir):
            if not os.path.isdir(targetDir):
                raise exceptions.OutputCreateDirectory(
                    "Specified target is a file, not a directory"
                )
        if not os.path.exists(targetDir):
            try:
                os.makedirs(targetDir)
            except OSError as e:
                raise exceptions.OutputCreateDirectory(
                    f"Failed to create target directory: [{e}]"
                )

    def open(
        self,
        targetFile: str,
        mode: Union[Literal["w"], Literal["wb"], Literal["rb"], Literal["r"]] = "wb",
    ) -> IO[Any]:
        """Open target file"""
        try:
            # pylint: disable=unspecified-encoding,consider-using-with
            self.fileHandle = builtins.open(targetFile, mode)
            return self.fileHandle
        except OSError as e:
            raise exceptions.OutputOpenException(
                f"Opening target file [{targetFile}] failed: {e}"
            ) from e

    def write(self, data: bytes) -> int:
        """Write wrapper"""
        self.chksum = zlib.adler32(data, self.chksum)
        written = self.fileHandle.write(data)
        assert written == len(data)
        return written

    def read(self, size=-1) -> int:
        """Read wrapper"""
        return self.fileHandle.read(size)

    def flush(self) -> None:
        """Flush wrapper"""
        return self.fileHandle.flush()

    def truncate(self, size: int) -> None:
        """Truncate target file"""
        try:
            self.fileHandle.truncate(size)
            self.fileHandle.seek(0)
        except OSError as e:
            raise exceptions.OutputException(
                f"Failed to truncate target file: [{e}]"
            ) from e

    def close(self) -> None:
        """Close wrapper"""
        log.debug("Close file")
        self.fileHandle.close()

    def seek(self, tgt: int, whence: int = 0) -> int:
        """Seek wrapper"""
        return self.fileHandle.seek(tgt, whence)

    def checksum(self) -> int:
        """Return computed checksum"""
        cur = self.chksum
        self.chksum = 1
        return cur



#!/usr/bin/python3
"""
Copyright (C) 2023 Michael Ablassmeier <abi@grinser.de>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import os
import tempfile
import logging
from argparse import Namespace
from libvirtnbdbackup import output
from libvirtnbdbackup import common as lib
from libvirtnbdbackup.objects import DomainDisk
from libvirtnbdbackup.virt import xml
from libvirtnbdbackup.virt import disktype


def read(ConfigFile: str) -> str:
    """Read saved virtual machine config'"""
    try:
        return output.openfile(ConfigFile, "rb").read().decode()
    except:
        logging.error("Can't read config file: [%s]", ConfigFile)
        raise


def changeVolumePathes(args: Namespace, vmConfig: str) -> bytes:
    """In case a virtual machine is using the volume based notation to
    configure disks, parsing the disk list from the configuration will
    fail because the volume is not existent anymore. Modify the disk
    setting from volume to file based before continuing to detect the
    attached disks. (#280)
    """
    tree = xml.asTree(vmConfig)
    for disk in tree.xpath("devices/disk"):
        dev = disk.xpath("target")[0].get("dev")
        diskType = disk.get("type")
        if diskType == "volume":
            source = disk.xpath("source")[0]
            disk.set("type", "file")
            volume = source.get("volume")
            volume = os.path.join(args.output, volume)
            source.set("file", volume)
            for attr in ["pool", "volume"]:
                source.attrib.pop(attr, None)
            logging.warning(
                "Disk [%s]: is using volume notation, overriding setting to [%s]",
                dev,
                volume,
            )

    return xml.ElementTree.tostring(tree, encoding="utf8", method="xml")


def removeDisk(vmConfig: str, excluded) -> bytes:
    """Remove disk from config, in case it has been excluded
    from the backup."""
    tree = xml.asTree(vmConfig)
    logging.info("Removing excluded disk [%s] from vm config.", excluded)
    try:
        target = tree.xpath(f"devices/disk/target[@dev='{excluded}']")[0]
        disk = target.getparent()
        disk.getparent().remove(disk)
    except IndexError:
        logging.warning("Removing excluded disk from config failed: no object found.")

    return xml.ElementTree.tostring(tree, encoding="utf8", method="xml")


def removeUuid(vmConfig: str) -> bytes:
    """Remove the auto generated UUID from the config file to allow
    for restore into new name"""
    tree = xml.asTree(vmConfig)
    try:
        logging.info("Removing uuid setting from vm config.")
        uuid = tree.xpath("uuid")[0]
        tree.remove(uuid)
    except IndexError:
        pass

    return xml.ElementTree.tostring(tree, encoding="utf8", method="xml")


def setVMName(args: Namespace, vmConfig: str) -> bytes:
    """Change / set the VM name to be restored"""
    tree = xml.asTree(vmConfig)
    name = tree.xpath("name")[0]
    if args.name is None and not name.text.startswith("restore"):
        domainName = f"restore_{name.text}"
        logging.info("Change VM name from [%s] to [%s]", name.text, domainName)
        name.text = domainName
    else:
        logging.info("Set name from [%s] to [%s]", name.text, args.name)
        name.text = args.name

    return xml.ElementTree.tostring(tree, encoding="utf8", method="xml")


def adjust(
    args: Namespace, restoreDisk: DomainDisk, vmConfig: str, targetFile: str
) -> bytes:
    """Adjust virtual machine configuration after restoring. Changes
    the paths to the virtual machine disks and attempts to remove
    components excluded during restore."""
    tree = xml.asTree(vmConfig)
    for disk in tree.xpath("devices/disk"):
        dev = disk.xpath("target")[0].get("dev")
        logging.debug("Handling target device: [%s]", dev)

        device = disk.get("device")
        driver = disk.xpath("driver")[0].get("type")

        if disktype.Optical(device, dev):
            logging.info("Removing device [%s], type [%s] from vm config", dev, device)
            disk.getparent().remove(disk)
            continue

        if disktype.Raw(driver, device) and args.raw is False:
            logging.warning(
                "Removing raw disk [%s] from vm config, use --raw to copy as is.",
                dev,
            )
            disk.getparent().remove(disk)
            continue

        backingStore = disk.xpath("backingStore")
        if backingStore:
            logging.info("Removing existent backing store settings")
            disk.remove(backingStore[0])

        dataStore = disk.xpath("source/dataStore")
        if dataStore and dev == restoreDisk.target:
            originalFile = os.path.basename(
                disk.xpath("source/dataStore/source")[0].get("file")
            )
            abspath = os.path.join(
                os.path.abspath(os.path.dirname(targetFile)), originalFile
            )
            logging.info(
                "Adjusting dataStore setting for disk [%s] from [%s] to [%s]",
                restoreDisk.target,
                disk.xpath("source/dataStore/source")[0].get("file"),
                abspath,
            )
            disk.xpath("source/dataStore/source")[0].set("file", abspath)

        originalFile = disk.xpath("source")[0].get("file")
        if dev == restoreDisk.target:
            abspath = os.path.abspath(targetFile)
            logging.info(
                "Change target file for disk [%s] from [%s] to [%s]",
                restoreDisk.target,
                originalFile,
                abspath,
            )
            disk.xpath("source")[0].set("file", abspath)

    return xml.ElementTree.tostring(tree, encoding="utf8", method="xml")


def restore(
    args: Namespace,
    vmConfig: str,
    adjustedConfig: bytes,
    targetFileName: str,
) -> None:
    """Restore either original or adjusted vm configuration
    to new directory"""
    targetFile = os.path.join(args.output, os.path.basename(targetFileName))
    if args.adjust_config is True:
        if args.sshClient:
            with tempfile.NamedTemporaryFile(delete=True) as fh:
                fh.write(adjustedConfig)
                lib.copy(args, fh.name, targetFile)
        else:
            with output.openfile(targetFile, "wb") as cnf:
                cnf.write(adjustedConfig)
            logging.info("Adjusted config placed in: [%s]", targetFile)
        if args.define is False:
            logging.info(
                "Use 'virsh -c %s define %s' to define VM", args.uri, targetFile
            )
    else:
        lib.copy(args, vmConfig, targetFile)
        logging.info("Copied original vm config to [%s]", targetFile)
        logging.info("Note: virtual machine config must be adjusted manually.")



"""
Copyright (C) 2023  Michael Ablassmeier <abi@grinser.de>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import os
import string
import random
import logging
from argparse import Namespace
from typing import Any, Dict, List, Tuple, Union
import libvirt
from libvirtnbdbackup.objects import DomainDisk
from libvirtnbdbackup.virt.exceptions import (
    domainNotFound,
    connectionFailed,
    startBackupFailed,
)
from libvirtnbdbackup.virt import fs
from libvirtnbdbackup.virt import xml
from libvirtnbdbackup.virt import disktype


def libvirt_ignore(
    _ignore: None, _err: Tuple[int, int, str, int, str, str, None, int, int]
) -> None:
    """this is required so libvirt.py does not report errors to stderr
    which it does by default. Error messages are fetched accordingly
    using exceptions.
    """


libvirt.registerErrorHandler(f=libvirt_ignore, ctx=None)

log = logging.getLogger("virt")


class client:
    """Libvirt related functions"""

    def __init__(self, uri: Namespace) -> None:
        self._conn: libvirt.virConnect = None
        self.libvirtVersion: int
        self.remoteHost: str = ""
        self.uri = uri
        self._domObj = None

    def connect(self) -> bool:
        """Connect libvirtd"""
        self._conn = self._connect(self.uri)
        self.libvirtVersion = self._conn.getLibVersion()

        return True

    @staticmethod
    def _connectAuth(uri: str, user: str, password: str) -> libvirt.virConnect:
        """Use openAuth if connection for advanced SASL authentication mechanisms
        if username and password are set"""

        def _cred(credentials, user_data) -> int:
            for credential in credentials:
                if credential[0] == libvirt.VIR_CRED_AUTHNAME:
                    credential[4] = user_data[0]
                elif credential[0] == libvirt.VIR_CRED_PASSPHRASE:
                    credential[4] = user_data[1]

            return 0

        log.debug("Username: %s", user)
        log.debug("Password: %s", password)

        try:
            flags: List[Any] = [libvirt.VIR_CRED_AUTHNAME, libvirt.VIR_CRED_PASSPHRASE]
            auth: List[Any] = [flags]
            if user is not None and password is not None:
                user_data = [user, password]
                auth.append(_cred)
                auth.append(user_data)

            return libvirt.openAuth(uri, auth, 0)
        except libvirt.libvirtError as e:
            raise connectionFailed(e) from e

    @staticmethod
    def _connectOpen(uri: str) -> libvirt.virConnect:
        """Open connection with regular libvirt URI for local authentication
        without further authentication mechanisms required"""
        try:
            return libvirt.open(uri)
        except libvirt.libvirtError as e:
            if e.get_error_code() == 45:
                errmsg = f"{e}: --user and --password options for SASL authentication are required."
                raise connectionFailed(errmsg) from e
            raise connectionFailed(e) from e

    def _connect(self, args: Namespace) -> libvirt.virConnect:
        """return libvirt connection handle and check if connection
        is established to a remote host."""
        log.debug("Libvirt URI: [%s]", args.uri)

        if args.user and args.password:
            conn = self._connectAuth(args.uri, args.user, args.password)
        else:
            conn = self._connectOpen(args.uri)

        # Detect if we are connected to a remote libvirt daemon by
        # comparing the local and remote hostname. If qemu+ssh is
        # part of the libvirt URI, set the remote host as well.
        # This will spawn the NBD service for data transfer via
        # TCP socket instead of local socket file and related virtual
        # domain files will be copied via SFTP.
        if "qemu+ssh" in args.uri:
            remoteHostname = conn.getHostname()
            log.info(
                "Connected to remote host: [%s] libvirt version: [%s]",
                remoteHostname,
                conn.getLibVersion(),
            )
            self.remoteHost = remoteHostname

        return conn

    def close(self) -> None:
        """Disconnect"""
        log.debug("Close connection to libvirt.")
        self._conn.close()

    def getDomain(self, name: str) -> libvirt.virDomain:
        """Lookup domain"""
        try:
            return self._conn.lookupByName(name)
        except libvirt.libvirtError as e:
            raise domainNotFound(e) from e

    def refreshPool(self, path: str) -> None:
        """Check if specified path matches an existing
        storage pool and refresh its contents"""
        try:
            pool = self._conn.storagePoolLookupByTargetPath(path)
        except libvirt.libvirtError:
            log.warning(
                "Restore path [%s] seems not to be an libvirt managed pool, skipping refresh.",
                path,
            )
            return

        try:
            pool.refresh()
            log.info("Refreshed contents of libvirt pool [%s]", pool.name())
        except libvirt.libvirtError as e:
            log.warning("Failed to refresh libvirt pool [%s]: [%s]", pool.name(), e)

    @staticmethod
    def blockJobActive(domObj: libvirt.virDomain, disks: List[DomainDisk]) -> bool:
        """Check if there is already an active block job for this virtual
        machine, which might block"""
        for disk in disks:
            blockInfo = domObj.blockJobInfo(disk.target)
            if (
                blockInfo
                and blockInfo["type"] == libvirt.VIR_DOMAIN_BLOCK_JOB_TYPE_BACKUP
            ):
                log.debug("Running block jobs for disk [%s]", disk.target)
                log.debug(blockInfo)
                return True
        return False

    def hasIncrementalEnabled(self, domObj: libvirt.virDomain) -> bool:
        """Check if virtual machine has enabled required capabilities
        for incremental backup

        Libvirt version >= 7006000  have the feature enabled
        by default without the domain XML including the capability
        statement.
        """
        if self.libvirtVersion >= 7006000:
            return True

        tree = xml.asTree(domObj.XMLDesc(0))
        for target in tree.findall(
            "{http://libvirt.org/schemas/domain/qemu/1.0}capabilities"
        ):
            for cap in target.findall(
                "{http://libvirt.org/schemas/domain/qemu/1.0}add"
            ):
                if "incremental-backup" in cap.items()[0]:
                    return True

        return False

    def _get_flags(self) -> int:
        """Check if both local and remote libvirt versions support required
        additional flags for possible backup enhancements"""
        flags = 0

        try:
            if self.libvirtVersion >= 11010000:
                flags = libvirt.VIR_DOMAIN_BACKUP_BEGIN_PRESERVE_SHUTDOWN_DOMAIN
                log.info("Setting supported flag to prevent vm shutdown during backup.")
        except AttributeError:
            pass

        return flags

    @staticmethod
    def getDomainConfig(domObj: libvirt.virDomain) -> str:
        """Return Virtual Machine configuration as XML"""
        return domObj.XMLDesc(0)

    @staticmethod
    def startDomain(domObj: libvirt.virDomain) -> bool:
        """Start virtual machine in paused state to allow full / inc backup"""
        return domObj.createWithFlags(
            flags=libvirt.VIR_DOMAIN_START_PAUSED | libvirt.VIR_DOMAIN_START_AUTODESTROY
        )

    @staticmethod
    def domainAutoStart(domObj: libvirt.virDomain) -> None:
        """Mark virtual machine for autostart"""
        try:
            domObj.setAutostart(1)
            log.info("Setting autostart config for domain.")
        except libvirt.libvirtError as errmsg:
            log.warning("Failed to set autostart flag for domain: [%s]", errmsg)

    def defineDomain(self, vmConfig: bytes, autoStart: bool) -> bool:
        """Define domain based on restored config"""
        try:
            log.info("Redefining domain based on adjusted config.")
            domObj = self._conn.defineXMLFlags(vmConfig.decode(), 0)
            log.info("Successfully redefined domain [%s]", domObj.name())
        except libvirt.libvirtError as errmsg:
            log.error("Failed to define domain: [%s]", errmsg)
            return False

        if autoStart is True:
            self.domainAutoStart(domObj)

        return True

    def getDomainInfo(self, vmConfig: str) -> Dict[str, str]:
        """Return object with general vm information relevant
        for backup"""
        tree = xml.asTree(vmConfig)
        settings = {}

        for flag in ["loader", "nvram", "kernel", "initrd"]:
            try:
                settings[flag] = tree.find("os").find(flag).text
            except AttributeError as e:
                log.debug("No setting [%s] found: %s", flag, e)

        log.debug("Domain Info: [%s]", settings)
        return settings

    def getTPMDevice(self, vmConfig: str) -> bool:
        """Check if virtual machine has configured an emulated (swtpm based) TPM device"""
        tree = xml.asTree(vmConfig)
        device = tree.find("devices/tpm")
        if device is not None:
            tpm = device.xpath("backend")[0].get("type")
            return tpm == "emulator"

        return False

    @staticmethod
    def getBackingStores(disk: xml._Element) -> List[str]:
        """Get list of backing store files defined for disk, usually
        the case if virtual machine has external snapshots."""
        backingStoreFiles: List[str] = []
        backingStore = disk.find("backingStore")
        while backingStore is not None:
            backingStoreSource = backingStore.find("source")

            if backingStoreSource is not None:
                backingStoreFiles.append(backingStoreSource.get("file"))

            if backingStore.find("backingStore") is not None:
                backingStore = backingStore.find("backingStore")
            else:
                backingStore = None

        return backingStoreFiles

    def _getDiskPathByVolume(self, disk: xml._Element) -> Union[str, None]:
        """If virtual machine disk is configured via type='volume'
        get path to disk via appropriate libvirt functions,
        pool and volume setting are mandatory as by xml schema definition"""
        vol = disk.xpath("source")[0].get("volume")
        pool = disk.xpath("source")[0].get("pool")

        try:
            diskPool = self._conn.storagePoolLookupByName(pool)
            diskPath = diskPool.storageVolLookupByName(vol).path()
        except libvirt.libvirtError as errmsg:
            log.error("Failed to detect vm disk by volumes: [%s]", errmsg)
            return None

        return diskPath

    def _hint(self, dev: str):
        """Show hint about possibility to reconfigure virtual machine with raw
        devices to support incremental backups"""

        if self.libvirtVersion <= 1010000:
            return

        msg = (
            "Check README on how to adjust virtual machine configuration"
            f" to enable full support for disk: [{dev}]."
        )
        log.warning(msg)

    def getDomainDisks(  # pylint: disable=too-many-branches
        self, args: Namespace, vmConfig: str
    ) -> List[DomainDisk]:
        """Parse virtual machine configuration for disk devices, filter
        all unsupported or excluded devices
        """
        devices = []

        excludeList = None
        if args.exclude is not None:
            excludeList = args.exclude.split(",")

        for disk in xml.asTree(vmConfig).xpath("devices/disk"):
            discardOption = None
            dev = disk.xpath("target")[0].get("dev")
            device = disk.get("device")
            diskFormat = disk.xpath("driver")[0].get("type")
            discardOption = disk.xpath("driver")[0].get("discard")

            if excludeList is not None and dev in excludeList:
                log.warning("Excluding disk [%s] from backup as requested", dev)
                continue

            if args.include is not None and dev != args.include:
                log.info(
                    "Skipping disk: [%s] as requested: does not match disk [%s]",
                    dev,
                    args.include,
                )
                continue

            # skip cdrom/floppy devices
            if disktype.Optical(device, dev):
                continue

            diskPath = None
            diskType = disk.get("type")
            if diskType == "volume":
                log.debug("Disk [%s]: volume notation", dev)
                diskPath = self._getDiskPathByVolume(disk)
            elif diskType == "file":
                log.debug("Disk [%s]: file notation", dev)
                diskPath = disk.xpath("source")[0].get("file")
            elif diskType == "block":
                # Direct attached block devices can be qcow formatted.
                # Skip only if format != qcow2 (#264)
                if args.raw is False and disktype.Raw(diskFormat, dev):
                    self._hint(dev)
                    continue
                diskPath = disk.xpath("source")[0].get("dev")
            elif diskType == "network":
                log.error("Unsupported network disk type for disk [%s]", dev)
                self._hint(dev)
                continue
            else:
                log.error("Unable to detect disk volume type for disk [%s]", dev)
                continue

            if diskPath is None:
                log.error("Unable to detect disk source for disk [%s]", dev)
                continue

            # skip direct attached devices if no --raw option is enabled
            if args.raw is False and (
                disktype.Block(disk, dev)
                or disktype.Lun(device, dev)
                or disktype.Raw(diskFormat, dev)
            ):
                self._hint(dev)
                continue

            diskFileName = os.path.basename(diskPath)
            backingStoreFiles = self.getBackingStores(disk)

            devices.append(
                DomainDisk(
                    dev,
                    diskFormat,
                    diskFileName,
                    diskPath,
                    backingStoreFiles,
                    discardOption,
                )
            )

        log.debug("Device list: %s ", devices)
        return devices

    def _createBackupXml(self, args: Namespace, diskList) -> str:
        """Create XML file for starting an backup task using libvirt API."""
        top = xml.ElementTree.Element("domainbackup", {"mode": "pull"})
        if self.remoteHost == "":
            xml.ElementTree.SubElement(
                top, "server", {"transport": "unix", "socket": f"{args.socketfile}"}
            )
        else:
            listen = self.remoteHost
            tls = "no"
            if args.tls:
                tls = "yes"
            if args.nbd_ip != "":
                listen = args.nbd_ip
            xml.ElementTree.SubElement(
                top,
                "server",
                {"tls": f"{tls}", "name": f"{listen}", "port": f"{args.nbd_port}"},
            )

        disks = xml.ElementTree.SubElement(top, "disks")

        if args.cpt.parent != "":
            inc = xml.ElementTree.SubElement(top, "incremental")
            inc.text = args.cpt.parent

        for disk in diskList:
            scratchId = "".join(
                random.choices(string.ascii_uppercase + string.digits, k=5)
            )
            scratchFile = os.path.join(
                args.scratchdir, f"backup.{scratchId}.{disk.target}"
            )
            log.debug("Using scratch file: %s", scratchFile)
            dE = xml.ElementTree.SubElement(disks, "disk", {"name": disk.target})
            xml.ElementTree.SubElement(dE, "scratch", {"file": f"{scratchFile}"})

        return xml.indent(top)

    def _createCheckpointXml(
        self, diskList: List[Any], parentCheckpoint: str, checkpointName: str
    ) -> str:
        """Create valid checkpoint XML file which is passed to libvirt API"""
        top = xml.ElementTree.Element("domaincheckpoint")
        desc = xml.ElementTree.SubElement(top, "description")
        desc.text = "Backup checkpoint"
        name = xml.ElementTree.SubElement(top, "name")
        name.text = checkpointName
        if parentCheckpoint != "":
            pct = xml.ElementTree.SubElement(top, "parent")
            cptName = xml.ElementTree.SubElement(pct, "name")
            cptName.text = parentCheckpoint
        disks = xml.ElementTree.SubElement(top, "disks")
        for disk in diskList:
            # No persistent checkpoint will be created for raw disks,
            # because it is not supported. Backup will only be crash
            # consistent. If we would like to create a consistent
            # backup, we would have to create an snapshot for these
            # kind of disks, example:
            # virsh checkpoint-create-as vm4 --diskspec sdb
            # error: unsupported configuration:  \
            # checkpoint for disk sdb unsupported for storage type raw
            # See also:
            # https://lists.gnu.org/archive/html/qemu-devel/2021-03/msg07448.html
            if disk.format != "raw":
                xml.ElementTree.SubElement(disks, "disk", {"name": disk.target})

        return xml.indent(top)

    def startBackup(
        self,
        args: Namespace,
        domObj: libvirt.virDomain,
        diskList: List[Any],
    ) -> None:
        """Attempt to start pull based backup task using  XML description"""
        backupXml = self._createBackupXml(args, diskList)
        checkpointXml = None
        freezed = False
        flags = self._get_flags()

        # do not create checkpoint during copy/diff backup.
        # backup saves delta until the last checkpoint
        if args.level not in ("copy", "diff"):
            checkpointXml = self._createCheckpointXml(
                diskList, args.cpt.parent, args.cpt.name
            )
        freezed = fs.freeze(domObj, args.freeze_mountpoint)
        try:
            log.debug("Starting backup job via libvirt API.")
            domObj.backupBegin(backupXml, checkpointXml, flags)
            log.debug("Started backup job via libvirt API.")
        except libvirt.libvirtError as errmsg:
            code = errmsg.get_error_code()
            if code == libvirt.VIR_ERR_CHECKPOINT_INCONSISTENT:
                raise startBackupFailed(
                    "Bitmap inconsistency detected: please cleanup checkpoints using virsh "
                    f"and execute new full backup: {errmsg}"
                ) from errmsg
            raise startBackupFailed(f"Failed to start backup: [{errmsg}]") from errmsg
        except Exception as e:
            log.exception(e)
            raise startBackupFailed(
                f"Unknown exception during backup start: [{e}]"
            ) from e
        finally:
            # check if filesystem is freezed and thaw
            # in case creating checkpoint fails.
            if freezed is True:
                fs.thaw(domObj)

    @staticmethod
    def stopBackup(domObj: libvirt.virDomain) -> bool:
        """Cancel the backup task using job abort"""
        try:
            domObj.abortJob()
            return True
        except libvirt.libvirtError as err:
            log.warning("Failed to stop backup job: [%s]", err)
            return False



#!/usr/bin/python3
"""
Copyright (C) 2023  Michael Ablassmeier <abi@grinser.de>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import logging


class logCount(logging.Handler):
    """Custom log handler keeping track of issued log messages"""

    class LogType:
        """Log message type"""

        def __init__(self) -> None:
            self.warnings = 0
            self.errors = 0

    def __init__(self) -> None:
        super().__init__()
        self.count = self.LogType()

    def emit(self, record: logging.LogRecord) -> None:
        if record.levelno == logging.WARNING:
            self.count.warnings += 1
        if record.levelno in (logging.ERROR, logging.CRITICAL):
            self.count.errors += 1



"""
Copyright (C) 2023  Michael Ablassmeier <abi@grinser.de>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import logging
import tempfile
import subprocess
from typing import List, Tuple, Union

from libvirtnbdbackup.qemu.exceptions import (
    ProcessError,
)
from libvirtnbdbackup.output import openfile
from libvirtnbdbackup.objects import processInfo

log = logging.getLogger(__name__)


def _readlog(logFile: str, cmd: str) -> str:
    try:
        with openfile(logFile, "rb") as fh:
            return fh.read().decode().strip()
    except Exception as errmsg:
        log.exception(errmsg)
        raise ProcessError(
            f"Failed to execute [{cmd}]: Unable to get error message: {errmsg}"
        ) from errmsg


def _readpipe(p) -> Tuple[str, str]:
    out = p.stdout.read().decode().strip()
    err = p.stderr.read().decode().strip()
    return out, err


def run(cmdLine: List[str], pidFile: str = "", toPipe: bool = False) -> processInfo:
    """Execute passed command"""
    logFileName: str = ""
    logFile: Union[int, tempfile._TemporaryFileWrapper]
    if toPipe is True:
        logFile = subprocess.PIPE
    else:
        # pylint: disable=consider-using-with
        logFile = tempfile.NamedTemporaryFile(
            delete=False, prefix=cmdLine[0], suffix=".log"
        )
        logFileName = logFile.name

    log.debug("CMD: %s", " ".join(cmdLine))
    try:
        with subprocess.Popen(
            cmdLine,
            close_fds=True,
            stderr=logFile,
            stdout=logFile,
        ) as p:
            p.wait()
            log.debug("Return code: %s", p.returncode)
            err: str = ""
            out: str = ""
            if p.returncode != 0:
                log.error("CMD: %s", " ".join(cmdLine))
                log.debug("Read error messages from logfile")
                if toPipe is True:
                    out, err = _readpipe(p)
                else:
                    err = _readlog(logFileName, cmdLine[0])
                raise ProcessError(f"Unable to start [{cmdLine[0]}] error: [{err}]")

            if toPipe is True:
                out, err = _readpipe(p)

            if pidFile != "":
                realPid = int(_readlog(pidFile, ""))
            else:
                realPid = p.pid

            process = processInfo(realPid, logFileName, err, out, pidFile)
            log.debug("Started [%s] process: [%s]", cmdLine[0], process)
    except FileNotFoundError as e:
        raise ProcessError(e) from e

    return process



#!/usr/bin/python3
"""
Copyright (C) 2023  Michael Ablassmeier <abi@grinser.de>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import os
import logging
from argparse import Namespace
from typing import List, IO, Any
from time import sleep
from libvirtnbdbackup import common as lib
from libvirtnbdbackup import output


def wait(offset: int, replayDevice: IO[Any], timeout: int = 60) -> None:
    """Sometimes seeking the NBD device may not yet be possible after
    it has just been initialized. Wait until we can seek to the biggest
    offset without OS error before continuing"""
    for _ in range(timeout):
        try:
            replayDevice.seek(offset)
            return
        except OSError:
            sleep(1)
    raise output.exceptions.OutputException("Timeout during setting up device.")


def replay(dataRanges: List, args: Namespace) -> None:
    """Replay the changes from an incremental or differential
    backup file to the NBD device"""
    logging.info("Replaying changes from incremental backups")
    blockListInc = list(
        filter(
            lambda x: x["inc"] is True,
            dataRanges,
        )
    )
    dataSize = sum(extent["length"] for extent in blockListInc)
    progressBar = lib.progressBar(dataSize, "replaying..", args)

    with output.openfile(args.device, "wb") as replayDevice:
        wait(blockListInc[-1]["originalOffset"], replayDevice)
        for extent in blockListInc:
            if args.noprogress:
                logging.info(
                    "Replaying offset [%s] from [%s] original offset [%s]",
                    extent["offset"],
                    extent["file"],
                    extent["originalOffset"],
                )
            with output.openfile(os.path.abspath(extent["file"]), "rb") as replaySrc:
                replaySrc.seek(extent["offset"])
                logging.debug(
                    "Seek [%s], to [%s], currently at: [%s]",
                    replayDevice.name,
                    extent["originalOffset"],
                    replayDevice.tell(),
                )
                replayDevice.seek(extent["originalOffset"])
                replayDevice.write(replaySrc.read(extent["length"]))
            replayDevice.seek(0)
            replayDevice.flush()
            progressBar.update(extent["length"])

    progressBar.close()



"""
Exceptions
"""


class CheckpointException(Exception):
    """Base checkpoint Exception"""


class NoCheckpointsFound(CheckpointException):
    """Inc or differential backup attempted but
    no existing checkpoints are found."""


class RedefineCheckpointError(CheckpointException):
    """During redefining existing checkpoints after
    vm relocate, an error occurred"""


class ReadCheckpointsError(CheckpointException):
    """Can't read checkpoint file"""


class RemoveCheckpointError(CheckpointException):
    """During removal of existing checkpoints after
    an error occurred"""


class SaveCheckpointError(CheckpointException):
    """Unable to append checkpoint to checkpoint
    file"""


class ForeignCeckpointError(CheckpointException):
    """Checkpoint for vm found which was not created
    by virtnbdbackup"""


class BackupException(Exception):
    """Base backup Exception"""


class DiskBackupFailed(BackupException):
    """Backup of one disk failed"""


class DiskBackupWriterException(BackupException):
    """Opening the target file writer
    failed"""


class RestoreException(Exception):
    """Base restore Exception"""


class UntilCheckpointReached(RestoreException):
    """Base restore Exception"""


class RestoreError(RestoreException):
    """Base restore error Exception"""



"""
Copyright (C) 2023  Michael Ablassmeier <abi@grinser.de>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import logging
import libvirt

log = logging.getLogger("fs")


def freeze(domObj: libvirt.virDomain, mountpoints: None) -> bool:
    """Attempt to freeze domain filesystems using qemu guest agent"""
    state, _ = domObj.state()
    if state == libvirt.VIR_DOMAIN_PAUSED:
        log.info("Skip freezing filesystems: domain is in paused state")
        return False

    log.debug("Attempting to freeze filesystems.")
    try:
        if mountpoints is not None:
            frozen = domObj.fsFreeze(mountpoints.split(","))
        else:
            frozen = domObj.fsFreeze()
        log.info("Freezed [%s] filesystems.", frozen)
        return True
    except libvirt.libvirtError as errmsg:
        log.warning(errmsg)
        return False


def thaw(domObj: libvirt.virDomain) -> bool:
    """Thaw freezed filesystems"""
    log.debug("Attempting to thaw filesystems.")
    try:
        thawed = domObj.fsThaw()
        log.info("Thawed [%s] filesystems.", thawed)
        return True
    except libvirt.libvirtError as errmsg:
        log.warning(errmsg)
        return False



"""
Copyright (C) 2023  Michael Ablassmeier <abi@grinser.de>
Copyright (C) 2020 Red Hat, Inc.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from dataclasses import dataclass


@dataclass(frozen=True)
class SparseStreamTypes:
    # pylint: disable=too-many-instance-attributes
    """Sparse stream format

    Extended format based on the examples provided by the
    ovirt-imageio project:

    https://github.com/oVirt/ovirt-imageio/tree/master/examples

    META:   start of meta information header
    DATA:   data block marker
    ZERO:   zero block marker
    STOP:   stop block marker
    TERM:   termination identifier
    FRAME:  assembled frame
    FRAME_LEN: length of frame

    Stream format
    =============

    Stream is composed of one of more frames.

    Meta frame
    ----------
    Stream metadata, must be the first frame.

    "meta" space start length "\r\n" <json-payload> \r\n

    Metadata keys in the json payload:

    - virtual-size: image virtual size in bytes
    - data-size: number of bytes in data frames
    - date: ISO 8601 date string

    Data frame
    ----------
    The header is followed by length bytes and terminator.
    "data" space start length "\r\n" <length bytes> "\r\n"

    Zero frame
    ----------
    A zero extent, no payload.
    "zero" space start length "\r\n"

    Stop frame
    ----------
    Marks the end of the stream, no payload.
    "stop" space start length "\r\n"

    Regular stream Example
    -------
    meta 0000000000000000 0000000000000083\r\n
    {
        [.]]
    }\r\n
    data 0000000000000000 00000000000100000\r\n
    <1 MiB bytes>\r\n
    zero 0000000000100000 00000000040000000\r\n
    data 0000000040100000 00000000000001000\r\n
    <4096 bytes>\r\n
    stop 0000000000000000 00000000000000000\r\n


    Compressed stream:
    -------
    Ends with compression marker:
    stop 0000000000000000 00000000000000000\r\n
    <json payload with compressed block sizes>\r\n
    comp 0000000000000000 00000000000000010\r\n
    """

    META: bytes = b"meta"
    DATA: bytes = b"data"
    COMP: bytes = b"comp"
    ZERO: bytes = b"zero"
    STOP: bytes = b"stop"
    TERM: bytes = b"\r\n"
    FRAME: bytes = b"%s %016x %016x" + TERM
    FRAME_LEN: int = len(FRAME % (STOP, 0, 0))



#!/usr/bin/python3
"""
Copyright (C) 2023 Michael Ablassmeier <abi@grinser.de>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
from typing import Dict
from libvirtnbdbackup import common as lib
from libvirtnbdbackup.sparsestream import streamer
from libvirtnbdbackup.sparsestream.exceptions import StreamFormatException
from libvirtnbdbackup.exceptions import RestoreError
from libvirtnbdbackup.output.exceptions import OutputException


def get(diskFile: str, stream: streamer.SparseStream) -> Dict[str, str]:
    """Read header from data file"""
    try:
        return lib.dumpMetaData(diskFile, stream)
    except StreamFormatException as errmsg:
        raise RestoreError(
            f"Reading metadata from [{diskFile}] failed: [{errmsg}]"
        ) from errmsg
    except OutputException as errmsg:
        raise RestoreError(
            f"Reading data file [{diskFile}] failed: [{errmsg}]"
        ) from errmsg
